__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import os
import re
import shutil

from beangulp import utils
from beangulp.exceptions import Error


# Template for the documents archival name.
FILENAME = '{date:%Y-%m-%d}.{name}'


def filepath(importer, filepath: str) -> str:
    """Compute filing path for a document.

    The path mirrors the structure of the accounts associated to the
    documents and with a file name composed by the document date and
    document name returned by the importer.

    Args:
      importer: The importer instance to handle the document.
      filepath: Filesystem path to the document.

    Returns:
      Filing tree location where to store the document.

    Raises:
      beangulp.exceptions.Error: The canonical file name returned by
      the importer contains a path separator charachter or seems to
      contain a date.

    """
    # Get the account corresponding to the file.
    account = importer.account(filepath)
    filename = importer.filename(filepath) or os.path.basename(filepath)
    date = importer.date(filepath) or utils.getmdate(filepath)

    # The returned filename cannot contain the file path separator character.
    if os.sep in filename:
        raise Error("The filename contains path separator character.")

    if re.match(r'\d\d\d\d-\d\d-\d\d\.', filename):
        raise Error("The filename contains what looks like a date.")

    # Prepend account directory and date prefix.
    filename = os.path.join(account.replace(':', os.sep), FILENAME.format(date=date, name=filename))

    return filename


def move(src: str, dst: str):
    """Move a file, potentially across devices.

    The destination direcory, and all intermediate path segments, are
    created if they do not exist. The move is performed with the
    shutil.move() function. See the documentation of this function for
    details of the semantic.

    Args:
      src: Filesystem path of the file to move.
      dst: Desitnation filesytem path. For the creation of the
        destination directory, this is assumed to be a file path and
        not a directory, namely the directory structure up to only
        path.dirname(dst) is created.

    """
    # Create missing directories.
    os.makedirs(os.path.dirname(dst), exist_ok=True)

    # Copy the file to its new name: use shutil.move() instead of
    # os.rename() to support moving across filesystems.
    shutil.move(src, dst)



import datetime
import os
import unittest

from os import path
from unittest import mock

from beangulp import archive
from beangulp import exceptions
from beangulp import tests


class TestFilepath(unittest.TestCase):

    def setUp(self):
        self.importer = tests.utils.Importer(None, 'Assets:Tests', None)

    def test_filepath(self):
        importer = mock.MagicMock(wraps=self.importer)
        importer.filename.return_value = 'foo.csv'
        filepath = archive.filepath(importer, path.abspath('test.pdf'))
        self.assertEqual(filepath, 'Assets/Tests/1970-01-01.foo.csv')

    def test_filepath_no_filename(self):
        filepath = archive.filepath(self.importer, path.abspath('test.pdf'))
        self.assertEqual(filepath, 'Assets/Tests/1970-01-01.test.pdf')

    def test_filepath_no_date(self):
        importer = mock.MagicMock(wraps=self.importer)
        importer.date.return_value = None
        with mock.patch('beangulp.archive.utils.getmdate',
                        return_value=datetime.datetime.fromtimestamp(
                            0, datetime.timezone.utc)):
            filepath = archive.filepath(importer, path.abspath('test.pdf'))
        self.assertEqual(filepath, 'Assets/Tests/1970-01-01.test.pdf')

    def test_filepath_sep_in_name(self):
        importer = mock.MagicMock(wraps=self.importer)
        importer.filename.return_value = f'dir{os.sep:}name.pdf'
        with self.assertRaises(exceptions.Error) as ex:
            archive.filepath(importer, path.abspath('test.pdf'))
        self.assertRegex(ex.exception.message, r'contains path separator')

    def test_filepath_date_in_name(self):
        importer = mock.MagicMock(wraps=self.importer)
        importer.filename.return_value = '1970-01-03.name.pdf'
        with self.assertRaises(exceptions.Error) as ex:
            archive.filepath(importer, path.abspath('test.pdf'))
        self.assertRegex(ex.exception.message, r'contains [\w\s]+ date')



import datetime
import decimal
import unittest
import petl

from beancount.parser import cmptest
from beangulp import petl_utils


class TestPetlUtils(cmptest.TestCase, unittest.TestCase):

    def test_table_to_directives_minimal(self):
        table = (
            petl.wrap([('date', 'account', 'amount'),
                       ('2021-02-15', 'Assets:Checking', '4.56'),
                       ('2021-02-16', 'Assets:Savings', '107.89')])
            .convert('date', lambda x: datetime.datetime.strptime(x, '%Y-%m-%d').date())
            .convert('amount', decimal.Decimal))
        transactions = petl_utils.table_to_directives(table)
        self.assertEqualEntries("""

          2021-02-15 *
            Assets:Checking  4.56 USD

          2021-02-16 *
            Assets:Savings  107.89 USD

        """, transactions)

    def test_table_to_directives_all(self):
        table = (
            petl.wrap([
                ('date', 'payee', 'narration', 'account', 'amount', 'balance'),
                ('2021-02-15', 'TheStore', 'BuyingSomething', 'Liabilities:Credit',
                 '-4.56', '-73330.00'),
            ])
            .convert('date', lambda x: datetime.datetime.strptime(x, '%Y-%m-%d').date())
            .convert(['amount', 'balance'], decimal.Decimal))
        transactions = petl_utils.table_to_directives(table)
        self.assertEqualEntries("""

          2021-02-15 * "TheStore" "BuyingSomething"
            Liabilities:Credit  -4.56 USD

          2021-02-16 balance Liabilities:Credit   -73330.00 USD

        """, transactions)



#!/usr/bin/env python3

from importers import acme
from importers import csvbank
from importers import ofx
from importers import utrade

from beancount.core import data
import beangulp


importers = [
    utrade.Importer("USD",
                    "Assets:US:UTrade",
                    "Assets:US:UTrade:Cash",
                    "Income:US:UTrade:{}:Dividend",
                    "Income:US:UTrade:{}:Gains",
                    "Expenses:Financial:Fees",
                    "Assets:US:BofA:Checking"),

    ofx.Importer("379700001111222",
                 "Liabilities:US:CreditCard",
                 "bofa"),

    acme.Importer("Assets:US:ACMEBank"),

    csvbank.Importer("Assets:US:CSVBank", "USD"),
]


def clean_up_descriptions(extracted_entries):
    """Example filter function; clean up cruft from narrations.

    Args:
      extracted_entries: A list of directives.
    Returns:
      A new list of directives with possibly modified payees and narration
      fields.
    """
    clean_entries = []
    for entry in extracted_entries:
        if isinstance(entry, data.Transaction):
            if entry.narration and " / " in entry.narration:
                left_part, _ = entry.narration.split(" / ")
                entry = entry._replace(narration=left_part)
            if entry.payee and " / " in entry.payee:
                left_part, _ = entry.payee.split(" / ")
                entry = entry._replace(payee=left_part)
        clean_entries.append(entry)
    return clean_entries


def process_extracted_entries(extracted_entries_list, ledger_entries):
    """Example filter function; clean up cruft from narrations.

    Args:
      extracted_entries_list: A list of (filename, entries) pairs, where
        'entries' are the directives extract from 'filename'.
      ledger_entries: If provided, a list of directives from the existing
        ledger of the user. This is non-None if the user provided their
        ledger file as an option.
    Returns:
      A possibly different version of extracted_entries_list, a list of
      (filename, entries), to be printed.
    """
    return [(filename, clean_up_descriptions(entries), account, importer)
            for filename, entries, account, importer in extracted_entries_list]


hooks = [process_extracted_entries]


if __name__ == '__main__':
    ingest = beangulp.Ingest(importers, hooks)
    ingest()



import fnmatch
from datetime import date

from beancount.parser import parser

from beangulp import importer
from beangulp import mimetypes


class Importer(importer.Importer):
    def __init__(self, name, account, mimetype):
        self._name = name
        self._account = account
        self._mimetype = mimetype

    @property
    def name(self):
        return self._name

    def identify(self, filepath):
        mimetype, encoding = mimetypes.guess_type(filepath, False)
        return mimetype == self._mimetype

    def date(self, filepath):
        return date(1970, 1, 1)

    def account(self, filepath):
        return self._account

    def filename(self, filepath):
        return None

    def extract(self, filepath, existing):
        return []

    def deduplicate(self, entries, existing):
        # Opt-out from the default deduplication mechanism.
        return entries


class IdentityImporter(importer.Importer):
    def __init__(self, name, account, match):
        self._name = name
        self._account = account
        self._match = match

    @property
    def name(self):
        return self._name

    def identify(self, filepath):
        return fnmatch.fnmatch(filepath, self._match)

    def date(self, filepath):
        return date(1970, 1, 1)

    def account(self, filepath):
        return self._account

    def filename(self, filepath):
        return None

    def extract(self, filepath, existing):
        entries, errors, options = parser.parse_file(filepath)
        return entries



import dateutil.parser


def parse_date(string, parse_kwargs_dict=None):
    """Parse arbitrary strings to dates.

    This function is intended to support liberal inputs, so that we can use it
    in accepting user-specified dates on command-line scripts.

    Args:
      string: A string to parse.
      parse_kwargs_dict: Dict of kwargs to pass to dateutil parser.
    Returns:
      A datetime.date object.
    """
    # At the moment, rely on the most excellent dateutil.
    if parse_kwargs_dict is None:
        parse_kwargs_dict = {}
    return dateutil.parser.parse(string, **parse_kwargs_dict).date()



"""Implementation of testing and generate functionality."""

from os import path
from typing import Callable, List, Optional, TextIO, Tuple, Union
import datetime
import difflib
import io
import os
import sys
import warnings

import click

from beancount.parser import printer
from beancount.core import data

import beangulp
from beangulp.importer import Importer, ImporterProtocol
from beangulp import extract
from beangulp import utils


def write_expected(outfile: TextIO,
                   account: data.Account,
                   date: Optional[datetime.date],
                   name: Optional[str],
                   entries: data.Entries):
    """Produce the expected output file.

    Args:
      outfile: The file object where to write.
      account: The account name produced by the importer.
      date: The date of the downloads file, produced by the importer.
      name: The filename for filing, produced by the importer.
      entries: The list of entries extracted by the importer.
    """
    date = date.isoformat() if date else ''
    name = name or ''
    print(f';; Account: {account}', file=outfile)
    print(f';; Date: {date}', file=outfile)
    print(f';; Name: {name}', file=outfile)
    printer.print_entries(entries, file=outfile)


def write_expected_file(filepath: str, *data, force: bool = False):
    """Writes out the expected file."""
    mode = 'w' if force else 'x'
    with open(filepath, mode) as expfile:
        write_expected(expfile, *data)


def compare_expected(filepath: str, *data) -> List[str]:
    """Compare the expected file with extracted data."""
    with io.StringIO() as buffer:
        write_expected(buffer, *data)
        # rewind to the beginning of the stream
        buffer.seek(0)
        lines_imported = buffer.readlines()

    with open(filepath, 'r') as infile:
        lines_expected = infile.readlines()

    diff = difflib.unified_diff(lines_expected, lines_imported,
                                tofile='expected.beancount',
                                fromfile='imported.beancount')
    return list(diff)


def run_importer(importer: Importer,
                 document: str) -> Tuple[data.Account,
                                         Optional[datetime.date],
                                         Optional[str],
                                         data.Entries]:
    """Run the various importer methods on the given cached file."""
    account = importer.account(document)
    date = importer.date(document)
    name = importer.filename(document)
    entries = extract.extract_from_file(importer, document, [])
    return account, date, name, entries


@click.command('test')
@click.argument('documents', nargs=-1,
                type=click.Path(exists=True, resolve_path=True))
@click.option('--expected', '-e', metavar='DIR',
              type=click.Path(file_okay=False, resolve_path=True),
              help="Directory containing the expecrted output files.")
@click.option('--verbose', '-v', count=True,
              help="Enable verbose output.")
@click.option('--quiet', '-q', count=True,
              help="Suppress all output.")
@click.option('--failfast', '-x', is_flag=True,
              help="Stop at the first test failure.")
@click.pass_obj
def _test(ctx,
          documents: List[str],
          expected: str,
          verbose: int,
          quiet: int,
          failfast: bool):
    """Test the importer.

    Run the importer on all DOCUMENTS and verify that it produces the
    desired output.  The desired output is stored in Beancount ledger
    files with a header containing additional metadata and can be
    generated with the "generate" command.

    The name of the desired output files is derived appending a
    ".beancount" suffix to the name of the input files, and are
    searched in the same directory where the input document is
    located, unless a different location is specified through the
    "--expected DIR" option.

    DOCUMENTS can be files or directories.  Directories are walked and
    the importer is called on all files with names not ending in
    ".beancount".  All and only the documents for which a desired
    output file exists must be positively identify by the importer.

    """
    return _run(ctx, documents, expected, verbose, quiet, failfast=failfast)


@click.command('generate')
@click.argument('documents', nargs=-1,
                type=click.Path(exists=True, resolve_path=True))
@click.option('--expected', '-e', metavar='DIR',
              type=click.Path(file_okay=False, resolve_path=True),
              help="Directory containing the expecrted output files.")
@click.option('--verbose', '-v', count=True,
              help="Enable verbose output.")
@click.option('--quiet', '-q', count=True,
              help="Suppress all output.")
@click.option('--force', '-f', is_flag=True,
              help='Alow to overwrite existing files.')
@click.pass_obj
def _generate(ctx,
              documents: List[str],
              expected: str,
              verbose: int,
              quiet: int,
              force: bool):
    """Generate expected files for tests.

    Run the importer on all DOCUMENTS and save the import results in
    Beancount ledger files with an header containing additional
    metadata that can be used to as regression tests for the importer.

    The name of the desired output files is derived appending a
    ".beancount" suffix to the name of the input files, and are
    written in the same directory where the input document is located,
    unless a different location is specified through the "--expected
    DIR" option.

    DOCUMENTS can be files or directories.  Directories are walked and
    the importer is called on all files with names not ending in
    ".beancount".

    """
    return _run(ctx, documents, expected, verbose, quiet, generate=True, force=force)


def _run(ctx,
         documents: List[str],
         expected: str,
         verbose: int,
         quiet: int,
         generate: bool = False,
         failfast: bool = False,
         force: bool = False):
    """Implement the test and generate commands."""

    assert len(ctx.importers) == 1
    importer = ctx.importers[0]

    verbosity = verbose - quiet
    log = utils.logger(verbosity)
    failures = 0

    for doc in utils.walk(documents):
        if doc.endswith('.beancount'):
            continue

        # Unless verbose mode is enabled, do not output a newline so
        # the test result is printed on the same line as the test
        # document filename.
        log(f'* {doc}', nl=verbosity > 0)

        # Compute the path to the expected output file.
        expected_filename = f"{doc}.beancount"
        if expected:
            expected_filename = path.join(expected, path.basename(expected_filename))

        # Run the importer's identify() method.
        if importer.identify(doc):
            account, date, name, entries = run_importer(importer, doc)
            log(f'  {expected_filename}', 1)
            if account is None:
                failures += 1
                log('  ERROR', fg='red')
                log('  ValueError: account() should not return None')
                continue
            log('  {}/{:%Y-%m-%d}-{}'.format(
                account.replace(":", "/"),
                date or utils.getmdate(doc),
                name or path.basename(doc)), 1)
            if generate:
                try:
                    write_expected_file(expected_filename, account, date, name, entries,
                        force=force)
                except FileExistsError as ex:
                    failures += 1
                    log('  ERROR', fg='red')
                    log('  FileExistsError: {}'.format(ex.filename))
                    continue
                log('  OK', fg='green')
                continue
            try:
                diff = compare_expected(expected_filename, account, date, name, entries)
            except FileNotFoundError:
                # The importer has positively identified a document
                # for which there is no expecred output file.
                failures += 1
                log('  ERROR', fg='red')
                log('  ExpectedOutputFileNotFound')
                continue
            if diff:
                # Test failure. Log an error.
                failures += 1
                log('  ERROR', fg='red')
                if verbosity >= 0:
                    sys.stdout.writelines(diff)
                    sys.stdout.write(os.linesep)
                    continue
            log('  PASSED', fg='green')

        elif path.exists(expected_filename):
            # The importer has not identified a document it should have.
            failures += 1
            log('  ERROR', fg='red')
            log('  DocumentNotIdentified')

        else:
            # Ignore files that are not positively identified by the
            # importer and for which there is no expected output file.
            log('  IGNORED')

        if failfast and failures:
            break

    if failures:
        sys.exit(1)


def wrap(importer: Union[Importer, ImporterProtocol]) -> Callable[[], None]:
    """Wrap a single importer for ingestion."""
    cli = beangulp.Ingest([importer]).cli
    cli.help = importer.__doc__
    cli.add_command(_test)
    cli.add_command(_generate)
    return cli


def main(importer: Union[Importer, ImporterProtocol]):
    """Call main program on a single importer. This is the main entry point."""
    if not sys.warnoptions:
        # Even if DeprecationWarnings are ignored by default print
        # them anyway unless other warnings settings are specified by
        # the -W Python command line flag.
        warnings.simplefilter('default')
    main = wrap(importer)
    main()



"""Mixin to add support for configuring importers with multiple accounts.

This importer implements some simple common functionality to create importers
which accept a long number of account names or regular expressions on the set of
account names. This is inspired by functionality in the importers in the
previous iteration of the ingest code, which used to be its own project.
"""
__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import logging


class ConfigImporterMixin:
    """A mixin class which supports configuration of account names.

    Mix this into the implementation of a importer.ImporterProtocol.
    """

    # A dict of required configuration variables to their documentation string.
    # This declares the list of options required for the importer to be provided
    # with, and their particular semantics.
    REQUIRED_CONFIG = {
        'FILE'       : 'Account for filing imported files to',
    }

    def __init__(self, config):
        """Provide a list of accounts and regexps as configuration to the importer.

        Args:
          config: A dict of configuration accounts, that must match the values
            declared in the class' REQUIRED_CONFIG.
        """
        super().__init__()

        # Check that the required configuration values are present.
        assert isinstance(config, dict), "Configuration must be a dict type"
        if not self._verify_config(config):
            raise ValueError("Invalid config {}, requires {}".format(
                config, self.REQUIRED_CONFIG))
        self.config = config

    def _verify_config(self, config):
        """Check the configuration account provided by the user against the accounts
        required by the source importer. Just to make sure.

        Args:
          config: A config dict of actual values on an importer.
          required_config: A dict of declarations of required values.
        Returns:
          A boolean, True on success.
        """
        provided_options = set(config)
        required_options = set(self.REQUIRED_CONFIG)

        success = True
        for option in (required_options - provided_options):
            logging.error("Missing value from user configuration for importer %s: %s",
                          self.name(), option)
            success = False

        for option in (provided_options - required_options):
            logging.error("Unknown value in user configuration for importer %s: %s",
                          self.name(), option)
            success = False

        # Note: Here we could validate account names further by looking them up
        # from the existing ledger of entries. This is under consideration.
        return success

    def file_account(self, _):
        return self.config['FILE']



import abc
import csv
import datetime
import decimal
from enum import Enum
import re

from collections import defaultdict
from itertools import islice
from beancount.core import data

import beangulp

EMPTY = frozenset()


def _resolve(spec, names):
    """Resolve column specification into column index.

    Args:
      spec: Column name or index.
      names: A dict mapping column names to column indices.

    Returns:
      Column index.
    """
    if isinstance(spec, int):
        return spec
    if names is None:
        raise KeyError(f'Column {spec!r} cannot be found in file without column names')
    col = names.get(spec)
    if col is None:
        cols = ', '.join(repr(name) for name in names.keys())
        raise KeyError(f'Cannot find column {spec!r} in column names: {cols}')
    return col


class Column:
    """Field descriptor.

    Args:
      name: Column name or index.
      default: Value to return if the field is empty.
    """

    def __init__(self, *names, default=None):
        self.names = names
        self.default = default

    def __repr__(self):
        names = ', '.join(repr(i) for i in self.names)
        return f'{self.__class__.__module__}.{self.__class__.__name__}({names})'

    def getter(self, names):
        """Generate an attribute accessor for the column specification.

        The returned function is a getitem-like function that takes a
        tuple as argument and returns the value of the field described
        by the column specificaion. The function returns the field
        default value if all columns are empty. The value is parsed
        with the column parser function.

        Args:
          names: A dict mapping column names to column indices.

        Returns:
          An accessor function.

        """
        idxs = [_resolve(x, names) for x in self.names]
        def func(obj):
            value = tuple(obj[i] for i in idxs)
            if not all(value) and self.default:
                return self.default
            return self.parse(*value)
        return func

    def parse(self, value):
        """Parse column value.

        Args:
          value: Field string value obtained from the CSV reader.

        Returns:
          Field parsed value.

        """
        return value.strip()


class Columns(Column):
    """Specialized Column for multiple column fields.

    Args:
      name: Column names or indexes.
      sep: Separator to use to join columns.

    """
    def __init__(self, *names, sep=' '):
        super().__init__(*names)
        self.sep = sep

    def parse(self, *values):
        return self.sep.join(val.strip() for val in values if val)


class Date(Column):
    """Specialized Column descriptor for date fields.

    Parse strings into datetime.date objects accordingly to the
    provided format specification. The format specification is the
    same understood by datetime.datetime.strptime().

    Args:
      name: Column name or index.
      frmt: Date format specification.

    """
    def __init__(self, name, frmt='%Y-%m-%d'):
        super().__init__(name)
        self.frmt = frmt

    def parse(self, value):
        return datetime.datetime.strptime(value.strip(), self.frmt).date()


class Amount(Column):
    """Specialized Column descriptor for decimal fields.

    Parse strings into decimal.Decimal objects. Optionally apply
    regexp substitutions before parsing the decimal number. This
    allows to normalize locale formatted decimal numbers into the
    format expected by decimal.Decimal().

    Args:
      name: Column name or index.
      subs: Dictionary mapping regular expression patterns to
        replacement strings. Substitutions are performed with
        re.sub() in the order they are specified.

    """

    def __init__(self, name, subs=None):
        super().__init__(name)
        self.subs = subs if subs is not None else {}

    def parse(self, value):
        for pattern, replacement in self.subs.items():
            value = re.sub(pattern, replacement, value)
        return decimal.Decimal(value)


# The CSV Importer class needs to inherit from beangulp.Importer which
# is an abstract base class having abc.ABCMeta as metaclass. To be
# able to do so out CSV metaclass need to be a sublcass of
# abc.ABCMeta.
class CSVMeta(abc.ABCMeta):
    """A metaclass that extracts column specifications from class members
    and stores them in a columns dictionary keyed by the member name."""
    def __new__(mcs, name, bases, dct):
        columns = {}
        others = {}
        for key, value in dct.items():
            if isinstance(value, Column):
                columns[key] = value
                continue
            others[key] = value
        others['columns'] = columns
        return super().__new__(mcs, name, bases, others)


class Order(Enum):
    ASCENDING = 1
    """Entries are listed in chronological order."""
    DESCENDING = 2
    """Entries are listed in reverse chronological order."""


class CSVReader(metaclass=CSVMeta):
    encoding = 'utf8'
    """File encoding."""
    skiplines = 0
    """Number of input lines to skip before starting processing."""
    names = True
    """Whether the data file contains a row with column names."""
    dialect = None
    """The CSV dialect used in the input file."""
    comments = '#'
    """Comment character."""
    order = None
    """Order of entries in the CSV file. If None the order will be inferred from the file content."""

    # This is populated by the CSVMeta metaclass.
    columns = {}

    def read(self, filepath):
        """Read CSV file according to class defined columns specification.

        Use the first rown in the data file to resolve columns
        specification. Yield namedtuple-like objects with attribute
        accessors to access the data fields as defined by the class
        columns specification.

        Args:
          filepath: Filesystem path to the input file.

        Yields:
          Namedtuple-like objects.

        """

        with open(filepath, encoding=self.encoding) as fd:
            # Skip header lines.
            lines = islice(fd, self.skiplines, None)

            # Filter out comment lines.
            if self.comments:
                lines = filter(lambda x: not x.startswith(self.comments), lines)

            reader = csv.reader(lines, dialect=self.dialect)

            # Map column names to column indices.
            names = None
            if self.names:
                headers = next(reader, None)
                if headers is None:
                    raise IndexError('The input file does not contain an header line')
                names = {name.strip(): index for index, name in enumerate(headers)}

            # Construct a class with attribute accessors for the
            # configured columns that works similarly to a namedtuple.
            attrs = {}
            for name, column in self.columns.items():
                attrs[name] = property(column.getter(names))
            row = type('Row', (tuple, ), attrs)

            # Return data rows.
            for x in reader:
                yield row(x)


class Importer(beangulp.Importer, CSVReader):
    """CSV files importer base class.

    This class provides the infrastructure and the basic functionality
    necessary for importing transactions and balance assertions from
    CSV files. To do anything useful it needs to be subclassed to add
    fields definitions. Fields are defined as class attributes of type
    Column.

    Args:
      account: Importer default account.
      currency: Importer default currency.
      flag: Importer default flag for new transactions.

    """
    def __init__(self, account, currency, flag='*'):
        self.importer_account = account
        self.currency = currency
        self.flag = flag

    def date(self, filepath):
        """Implement beangulp.Importer::date()"""
        return max(row.date for row in self.read(filepath))

    def account(self, filepath):
        """Implement beangulp.Importer::account()"""
        return self.importer_account

    def extract(self, filepath, existing):
        """Implement beangulp.Importer::extract()

        This methods costructs a transaction for each data row using
        the date, narration, and amount required fields and the flag,
        payee, account, currency, tag, link, balance optional fields.

        Transaction metadata is constructed with the metadata() method
        and the finalize() method is called on each transaction. These
        can be redefine in subclasses. For customization that cannot
        be implemented with these two extension points, consider
        basing the importer on the CSVReader class and implement
        tailored data processing in the extract() method.

        """

        entries = []
        balances = defaultdict(list)
        default_account = self.account(filepath)

        # Compute the line number of the first data line.
        offset = int(self.skiplines) + bool(self.names) + 1

        for lineno, row in enumerate(self.read(filepath), offset):
            # Skip empty lines.
            if not row:
                continue

            tag = getattr(row, 'tag', None)
            tags = {tag} if tag else EMPTY

            link = getattr(row, 'link', None)
            links = {link} if link else EMPTY

            # This looks like an exercise in defensive programming
            # gone too far, but we do not want to depend on any field
            # being defined other than the essential ones.
            flag = getattr(row, 'flag', self.flag)
            payee = getattr(row, 'payee', None)
            account = getattr(row, 'account', default_account)
            currency = getattr(row, 'currency', self.currency)
            units = data.Amount(row.amount, currency)

            # Create a transaction.
            txn = data.Transaction(self.metadata(filepath, lineno, row),
                                   row.date, flag, payee, row.narration, tags, links, [
                                       data.Posting(account, units, None, None, None, None),
                                   ])

            # Apply user processing to the transaction.
            txn = self.finalize(txn, row)
            if txn is None:
                continue

            # Add the transaction to the output list.
            entries.append(txn)

            # Add balance to balances list.
            balance = getattr(row, 'balance', None)
            if balance is not None:
                date = row.date + datetime.timedelta(days=1)
                units = data.Amount(balance, currency)
                meta = data.new_metadata(filepath, lineno)
                balances[currency].append(data.Balance(meta, date, account, units, None, None))

        if not entries:
            return []

        if self.order is None:
            self.order = Order.ASCENDING if entries[0].date <= entries[-1].date else Order.DESCENDING

        # Reverse the list if the file is in descending order.
        if self.order is Order.DESCENDING:
            entries.reverse()

        # Append balances.
        for currency, balances in balances.items():
            entries.append(balances[-1 if self.order is Order.ASCENDING else 0])

        return entries

    def metadata(self, filepath, lineno, row):
        """Build transaction metadata dictionary.

        This method can be extended to add customized metadata
        entries based on the content of the data row.

        Args:
          filepath: Path to the file being imported.
          lineno: Line number of the data being processed.
          row: The data row being processed.

        Returns:
          A metadata dictionary.

        """
        return data.new_metadata(filepath, lineno)

    def finalize(self, txn, row):
        """Post process the transaction.

        Returning None results in the transaction being discarded and
        in source row to do not contribute to the determination of the
        balances.

        Args:
          txn: The just build Transaction object.
          row: The data row being processed.

        Returns:
          A potentially extended or modified Transaction object or None.

        """
        return txn



import datetime
import unittest

from beangulp import date_utils


class TestDateUtils(unittest.TestCase):

    def test_parse_date(self):
        self.assertEqual(datetime.date(2021, 7, 4), date_utils.parse_date('2021-07-04'))



from os import path
from beangulp import mimetypes
from beangulp.importers import csvbase
from beangulp.testing import main


class Importer(csvbase.Importer):
    date = csvbase.Date('Posting Date', '%m/%d/%Y')
    narration = csvbase.Columns('Description', 'Check or Slip #', sep='; ')
    amount = csvbase.Amount('Amount')
    balance = csvbase.Amount('Balance')

    def identify(self, filepath):
        mimetype, encoding = mimetypes.guess_type(filepath)
        if mimetype != 'text/csv':
            return False
        with open(filepath) as fd:
            head = fd.read(1024)
        return head.startswith('Details,Posting Date,"Description",'
                               'Amount,Type,Balance,Check or Slip #,')

    def filename(self, filepath):
        return 'csvbank.' + path.basename(filepath)


if __name__ == '__main__':
    main(Importer('Assets:US:CSVBank', 'USD'))



"""Example importer for PDF statements from ACME Bank.

This importer identifies the file from its contents and only supports
filing, it cannot extract any transactions from the PDF conversion to
text.  This is common, and I figured I'd provide an example for how
this works.

"""
__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import re
import subprocess

from dateutil.parser import parse as parse_datetime

import beangulp
from beangulp import mimetypes
from beangulp.cache import cache
from beangulp.testing import main


@cache
def pdf_to_text(filename):
    """Convert a PDF document to a text equivalent."""
    r = subprocess.run(['pdftotext', filename, '-'],
                       stdout=subprocess.PIPE, check=True)
    return r.stdout.decode()


class Importer(beangulp.Importer):
    """An importer for ACME Bank PDF statements."""

    def __init__(self, account_filing):
        self.account_filing = account_filing

    def identify(self, filepath):
        mimetype, encoding = mimetypes.guess_type(filepath)
        if mimetype != 'application/pdf':
            return False

        # Look for some words in the PDF file to figure out if it's a statement
        # from ACME. The filename they provide (Statement.pdf) isn't useful.
        text = pdf_to_text(filepath)
        if text:
            return re.match('ACME Bank', text) is not None

    def filename(self, filepath):
        # Normalize the name to something meaningful.
        return 'acmebank.pdf'

    def account(self, filepath):
        return self.account_filing

    def date(self, filepath):
        # Get the actual statement's date from the contents of the file.
        text = pdf_to_text(filepath)
        match = re.search('Date: ([^\n]*)', text)
        if match:
            return parse_datetime(match.group(1)).date()


if __name__ == '__main__':
    importer = Importer("Assets:US:ACMEBank")
    main(importer)



__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

from beangulp.exceptions import Error


# A file size beyond which we will simply ignore the file. This is
# used to skip large files that are commonly present in a Downloads
# directory.
FILE_TOO_LARGE_THRESHOLD = 8*1024*1024


def identify(importers, filepath: str):
    """Identify the correct importer to handle a document.

    Args:
      importers: List of importer instances.
      filepath: Filesystem path to the document.

    Returns:
      The correct importer to handle the document or None if no importer
      matches the document.

    Raises:
      beangulp.exceptions.Error: More than one importer matched the file.

    """
    match = [importer for importer in importers if importer.identify(filepath)]
    if len(match) > 1:
        match = [f'  {importer.name}' for importer in match]
        raise Error('Document identified by more than one importer.', *match)

    return match[0] if match else None



import io
import textwrap
import unittest

from datetime import timedelta
from os import path
from unittest import mock

from beancount.parser import parser
from beangulp import extract
from beangulp import similar
from beangulp import tests


class TestExtract(unittest.TestCase):

    def setUp(self):
        self.importer = tests.utils.Importer(None, 'Assets:Tests', None)

    def test_extract_from_file_no_entries(self):
        entries = extract.extract_from_file(self.importer, path.abspath('test.csv'), [])
        self.assertEqual(entries, [])

    def test_extract_from_file(self):
        entries, errors, options = parser.parse_string(textwrap.dedent('''
            1970-01-03 * "Test"
              Assets:Tests  1.00 USD

            1970-01-01 * "Test"
              Assets:Tests  1.00 USD

            1970-01-02 * "Test"
              Assets:Tests  1.00 USD
            '''))

        importer = mock.MagicMock(wraps=self.importer)
        importer.extract.return_value = entries
        entries = extract.extract_from_file(importer, path.abspath('test.csv'), [])
        dates = [entry.date for entry in entries]
        self.assertSequenceEqual(dates, sorted(dates))

    def test_extract_from_file_ensure_sanity(self):
        entries, errors, options = parser.parse_string('''
            1970-01-01 * "Test"
              Assets:Tests  1.00 USD
            ''')

        # Break something.
        entries[-1] = entries[-1]._replace(narration=42)
        importer = mock.MagicMock(wraps=self.importer)
        importer.extract.return_value = entries
        with self.assertRaises(AssertionError):
            extract.extract_from_file(importer, path.abspath('test.csv'), [])


class TestDuplicates(unittest.TestCase):

    def test_mark_duplicate_entries(self):
        entries, error, options = parser.parse_string(textwrap.dedent('''
          1970-01-01 * "Test"
            Assets:Tests  10.00 USD

          1970-01-02 * "Test"
            Assets:Tests  20.00 USD
        '''))
        compare = similar.heuristic_comparator()
        extract.mark_duplicate_entries(entries, entries[:1], timedelta(days=2), compare)
        self.assertTrue(entries[0].meta[extract.DUPLICATE])
        self.assertNotIn(extract.DUPLICATE, entries[1].meta)


class TestPrint(unittest.TestCase):

    def test_print_extracted_entries(self):
        entries, error, options = parser.parse_string(textwrap.dedent('''
            1970-01-01 * "Test"
              Assets:Tests  10.00 USD'''))

        extracted = [
            ('/path/to/test.csv', entries, None, None),
            ('/path/to/empty.pdf', [], None, None),
        ]

        output = io.StringIO()
        extract.print_extracted_entries(extracted, output)

        self.assertEqual(output.getvalue(), textwrap.dedent('''\
            ;; -*- mode: beancount -*-

            **** /path/to/test.csv

            1970-01-01 * "Test"
              Assets:Tests  10.00 USD


            **** /path/to/empty.pdf


            '''))

    def test_print_extracted_entries_duplictes(self):
        entries, error, options = parser.parse_string(textwrap.dedent('''
            1970-01-01 * "Test"
              Assets:Tests  10.00 USD

            1970-01-01 * "Test"
              Assets:Tests  10.00 USD '''))

        # Mark the second entry as duplicate
        entries[1].meta[extract.DUPLICATE] = True

        extracted = [
            ('/path/to/test.csv', entries, None, None),
        ]

        output = io.StringIO()
        extract.print_extracted_entries(extracted, output)

        self.assertEqual(output.getvalue(), textwrap.dedent('''\
            ;; -*- mode: beancount -*-

            **** /path/to/test.csv

            1970-01-01 * "Test"
              Assets:Tests  10.00 USD

            ; 1970-01-01 * "Test"
            ;   Assets:Tests  10.00 USD


            '''))



__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import unittest

from beangulp import importer
from beangulp.importers import config


class SimpleTestImporter(config.ConfigImporterMixin,
                         importer.ImporterProtocol):
    REQUIRED_CONFIG = {
        'FILE'         : 'Account for filing',
        'asset_cash'   : 'Cash account',
        'asset_shares' : 'Positions',
        'dividend'     : 'Dividends',
        'fees'         : 'Fees',
    }


class TestConfigMixin(unittest.TestCase):

    def setUp(self):
        self.default_config = {key: 'Assets:Something'
                               for key in SimpleTestImporter.REQUIRED_CONFIG}

    def test_constructors(self):
        # Test invalid input type.
        with self.assertRaises(AssertionError):
            SimpleTestImporter('*')

        # Test a succeeding case.
        SimpleTestImporter(self.default_config)

    def test_file_account(self):
        config = self.default_config.copy()
        config['FILE'] = 'Assets:FilingAccount'
        importer = SimpleTestImporter(config)
        self.assertEqual('Assets:FilingAccount', importer.file_account(None))

    def test_invalid_missing(self):
        config = self.default_config.copy()
        del config['asset_cash']
        with self.assertRaises(ValueError):
            SimpleTestImporter(config)

    def test_invalid_extra(self):
        config = self.default_config.copy()
        config['asset_other'] = 'Assets:Other'
        with self.assertRaises(ValueError):
            SimpleTestImporter(config)



import unittest

from os import path

from beangulp import exceptions
from beangulp import identify
from beangulp import tests


class TestIdentify(unittest.TestCase):

    def test_identify(self):
        importers = [
            tests.utils.Importer('A', 'Assets:Tests', 'application/pdf'),
            tests.utils.Importer('B', 'Assets:Tests', 'text/csv'),
        ]

        # Pass an absolute path to identify() to make the cache code
        # used internally by the importers happy. This can go away
        # once FileMemo is removed from the importers interface.
        importer = identify.identify(importers, path.abspath('test.txt'))
        self.assertIsNone(importer)

        importer = identify.identify(importers, path.abspath('test.pdf'))
        self.assertEqual(importer.name, 'A')

        importer = identify.identify(importers, path.abspath('test.csv'))
        self.assertEqual(importer.name, 'B')

    def test_identify_collision(self):
        importers = [
            tests.utils.Importer('A', 'Assets:Tests', 'text/csv'),
            tests.utils.Importer('B', 'Assets:Tests', 'text/csv'),
        ]

        importer = identify.identify(importers, path.abspath('test.txt'))
        self.assertIsNone(importer)

        with self.assertRaises(exceptions.Error):
            importer = identify.identify(importers, path.abspath('test.csv'))



__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import tempfile
import textwrap
import unittest
import warnings

from pprint import pformat
from beancount.core import data
from beancount.parser import cmptest
from beancount.utils import test_utils
from beangulp.importers import csv


Col = csv.Col


class TestCSVFunctions(unittest.TestCase):

    def test_normalize_config__with_header(self):
        head = textwrap.dedent("""\
          Details,Posting Date,"Description",Amount,Type,Balance,Check or Slip #,
          DEBIT,3/18/2016,"Payment to Chafe card ending in 1234 03/18",-2680.89,ACCT_XFER,3409.86,,
          CREDIT,3/15/2016,"EMPLOYER INC    DIRECT DEP                 PPD ID: 1111111111",2590.73,ACH_CREDIT,6090.75,,
          DEBIT,3/14/2016,"INVESTMENT SEC   TRANSFER   A5144608        WEB ID: 1234456789",-150.00,ACH_DEBIT,3500.02,,
        """)
        iconfig, has_header = csv.normalize_config({Col.DATE: 'Posting Date'}, head)
        self.assertEqual({Col.DATE: 1}, iconfig)
        self.assertTrue(has_header)

        iconfig, _ = csv.normalize_config({Col.NARRATION: 'Description'}, head)
        self.assertEqual({Col.NARRATION: 2}, iconfig)

        iconfig, _ = csv.normalize_config({Col.DATE: 1,
                                           Col.NARRATION: 'Check or Slip #'},
                                          head)
        self.assertEqual({Col.DATE: 1, Col.NARRATION: 6}, iconfig)

        iconfig, _ = csv.normalize_config({Col.DATE: 1}, head)
        self.assertEqual({Col.DATE: 1}, iconfig)

    def test_normalize_config__with_skip_and_header(self):
        head = textwrap.dedent("""\
          Lorem ipsum dolor sit amet, consectetur adipiscing elit.
          Aliquam lorem erat, bibendum sed arcu at, tempor commodo tortor.
          Phasellus consectetur, nisl quis vestibulum ornare, mi velit imperdiet arcu, eu mattis nulla augue nec ex.

          Details,Posting Date,"Description",Amount,Type,Balance,Check or Slip #,
          DEBIT,3/18/2016,"Payment to Chafe card ending in 1234 03/18",-2680.89,ACCT_XFER,3409.86,,
          CREDIT,3/15/2016,"EMPLOYER INC    DIRECT DEP                 PPD ID: 1111111111",2590.73,ACH_CREDIT,6090.75,,
          DEBIT,3/14/2016,"INVESTMENT SEC   TRANSFER   A5144608        WEB ID: 1234456789",-150.00,ACH_DEBIT,3500.02,,
        """)
        iconfig, has_header = csv.normalize_config({Col.DATE: 'Posting Date'}, head,
                                                   skip_lines=4)
        self.assertEqual({Col.DATE: 1}, iconfig)
        self.assertTrue(has_header)

        iconfig, _ = csv.normalize_config({Col.NARRATION: 'Description'}, head,
                                          skip_lines=4)
        self.assertEqual({Col.NARRATION: 2}, iconfig)

        iconfig, _ = csv.normalize_config({Col.DATE: 1,
                                           Col.NARRATION: 'Check or Slip #'},
                                          head, skip_lines=4)
        self.assertEqual({Col.DATE: 1, Col.NARRATION: 6}, iconfig)

        iconfig, _ = csv.normalize_config({Col.DATE: 1}, head, skip_lines=4)
        self.assertEqual({Col.DATE: 1}, iconfig)

    def test_normalize_config__without_header(self):
        head = textwrap.dedent("""\
          DEBIT,3/18/2016,"Payment to Chafe card ending in 1234 03/18",-2680.89,ACCT_XFER,3409.86,,
          CREDIT,3/15/2016,"EMPLOYER INC    DIRECT DEP                 PPD ID: 1111111111",2590.73,ACH_CREDIT,6090.75,,
          DEBIT,3/14/2016,"INVESTMENT SEC   TRANSFER   A5144608        WEB ID: 1234456789",-150.00,ACH_DEBIT,3500.02,,
        """)
        iconfig, has_header = csv.normalize_config({Col.DATE: 1}, head)
        self.assertEqual({Col.DATE: 1}, iconfig)
        self.assertFalse(has_header)
        with self.assertRaises(ValueError):
            csv.normalize_config({Col.DATE: 'Posting Date'}, head)

    def test_normalize_config__with_skip_and_without_header(self):
        head = textwrap.dedent("""\
          Lorem ipsum dolor sit amet, consectetur adipiscing elit.
          Aliquam lorem erat, bibendum sed arcu at, tempor commodo tortor.
          Phasellus consectetur, nisl quis vestibulum ornare, mi velit imperdiet arcu, eu mattis nulla augue nec ex.

          DEBIT,3/18/2016,"Payment to Chafe card ending in 1234 03/18",-2680.89,ACCT_XFER,3409.86,,
          CREDIT,3/15/2016,"EMPLOYER INC    DIRECT DEP                 PPD ID: 1111111111",2590.73,ACH_CREDIT,6090.75,,
          DEBIT,3/14/2016,"INVESTMENT SEC   TRANSFER   A5144608        WEB ID: 1234456789",-150.00,ACH_DEBIT,3500.02,,
        """)
        iconfig, has_header = csv.normalize_config({Col.DATE: 1}, head, skip_lines=4)
        self.assertEqual({Col.DATE: 1}, iconfig)
        self.assertFalse(has_header)
        with self.assertRaises(ValueError):
            csv.normalize_config({Col.DATE: 'Posting Date'}, head, skip_lines=4)


class TestCSVImporter(cmptest.TestCase):

    def setUp(self):
        warnings.simplefilter('ignore', category=DeprecationWarning)

    @test_utils.docfile
    def test_column_types(self, filename):
        # pylint: disable=line-too-long
        """\
          Details,Posting Date,"Description",Amount,Type,Balance,Check or Slip #,
          DEBIT,3/18/2016,"Payment to Chafe card ending in 1234 03/18",-2680.89,ACCT_XFER,3409.86,,
          CREDIT,3/15/2016,"EMPLOYER INC    DIRECT DEP                 PPD ID: 1111111111",2590.73,ACH_CREDIT,6090.75,,
          DEBIT,3/14/2016,"INVESTMENT SEC   TRANSFER   A5144608        WEB ID: 1234456789",-150.00,ACH_DEBIT,3500.02,,
          DEBIT,3/6/2016,"ATM WITHDRAWAL                       001234  03/8888 DELANC",-60.00,ATM,3650.02,,
          CREDIT,3/5/2016,"CA STATE         NYSTTAXRFD                 PPD ID: 1111111111",110.00,ACH_CREDIT,3710.02,,
          DEBIT,3/4/2016,"BOOGLE           WALLET     US000NEI9T      WEB ID: C234567890",-1300.00,ACH_DEBIT,3600.02,,
        """

        importer = csv.CSVImporter({Col.DATE: 'Posting Date',
                                    Col.NARRATION1: 'Description',
                                    Col.NARRATION2: 'Check or Slip #',
                                    Col.AMOUNT: 'Amount',
                                    Col.BALANCE: 'Balance',
                                    Col.DRCR: 'Details'},
                                   'Assets:Bank',
                                   'USD',
                                   ('Details,Posting Date,"Description",Amount,'
                                    'Type,Balance,Check or Slip #,'),
                                   institution='chafe')
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

          2016-03-18 * "Payment to Chafe card ending in 1234 03/18"
            Assets:Bank  -2680.89 USD

          2016-03-15 * "EMPLOYER INC    DIRECT DEP                 PPD ID: 1111111111"
            Assets:Bank  2590.73 USD

          2016-03-14 * "INVESTMENT SEC   TRANSFER   A5144608        WEB ID: 1234456789"
            Assets:Bank  -150.00 USD

          2016-03-06 * "ATM WITHDRAWAL                       001234  03/8888 DELANC"
            Assets:Bank  -60.00 USD

          2016-03-05 * "CA STATE         NYSTTAXRFD                 PPD ID: 1111111111"
            Assets:Bank  110.00 USD

          2016-03-04 * "BOOGLE           WALLET     US000NEI9T      WEB ID: C234567890"
            Assets:Bank  -1300.00 USD

          2016-03-19 balance Assets:Bank                                     3409.86 USD

        """, entries)

    @test_utils.docfile
    def test_date_formats(self, filename):
        """\
          Posting,Description,Amount
          11/7/2016,A,2
          12/7/2016,B,3
          13/7/2016,C,4
        """
        importer = csv.CSVImporter({Col.DATE: 'Posting',
                                    Col.NARRATION: 'Description',
                                    Col.AMOUNT: 'Amount'},
                                   'Assets:Bank', 'EUR', [],
                                   dateutil_kwds={'dayfirst': True})
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

          2016-07-11 * "A"
            Assets:Bank  2 EUR

          2016-07-12 * "B"
            Assets:Bank  3 EUR

          2016-07-13 * "C"
            Assets:Bank  4 EUR

        """, entries)


    @test_utils.docfile
    def test_links(self, filename):
        """\
          Date,Description,Amount,Link
          2020-07-03,A,2,
          2020-07-03,B,3,123
        """
        importer = csv.CSVImporter({Col.DATE: 'Date',
                                    Col.NARRATION: 'Description',
                                    Col.AMOUNT: 'Amount',
                                    Col.REFERENCE_ID: 'Link'},
                                   'Assets:Bank', 'EUR', [])
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

          2020-07-03 * "A"
            Assets:Bank  2 EUR

          2020-07-03 * "B" ^123
            Assets:Bank  3 EUR
        """, entries)


    @test_utils.docfile
    def test_tags(self, filename):
        """\
          Date,Description,Amount,Tag
          2020-07-03,A,2,
          2020-07-03,B,3,foo
        """
        importer = csv.CSVImporter({Col.DATE: 'Date',
                                    Col.NARRATION: 'Description',
                                    Col.AMOUNT: 'Amount',
                                    Col.TAG: 'Tag'},
                                   'Assets:Bank', 'EUR', [])
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

          2020-07-03 * "A"
            Assets:Bank  2 EUR

          2020-07-03 * "B" #foo
            Assets:Bank  3 EUR
        """, entries)


    @test_utils.docfile
    def test_zero_balance_produces_assertion(self, filename):
        """\
          Details,Posting Date,"Description",Amount,Type,Balance,Check or Slip #,
          DEBIT,3/18/2016,"Payment to Chafe card ending in 1234 03/18",-2680.89,ACCT_XFER,0,,
        """
        importer = csv.CSVImporter({Col.DATE: 'Posting Date',
                                    Col.NARRATION1: 'Description',
                                    Col.NARRATION2: 'Check or Slip #',
                                    Col.AMOUNT: 'Amount',
                                    Col.BALANCE: 'Balance',
                                    Col.DRCR: 'Details'},
                                   'Assets:Bank',
                                   'USD',
                                   ('Details,Posting Date,"Description",Amount,'
                                    'Type,Balance,Check or Slip #,'),
                                   institution='chafe')
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

          2016-03-18 * "Payment to Chafe card ending in 1234 03/18"
            Assets:Bank  -2680.89 USD

          2016-03-19 balance Assets:Bank                                     0 USD

        """, entries)

    @test_utils.docfile
    def test_currency_and_balances_where_there_are_multiple_currency_transactions(self, filename):
        """\
          Posting Date,"Description",Amount,Currency,Balance
          3/18/2016,"1st Payment in GBP",-1.00,GBP,-1
          3/18/2016,"1st Payment in PLN",-1,PLN,-1
          3/18/2016,"1st Payment in ZAR",-1.0,ZAR,-1
          3/19/2016,"2nd Payment in GBP",-2,GBP,-3
          3/19/2016,"2nd Payment in Main Currency",-2.00,,-3
          3/20/2016,"3rd Payment in GBP",-3,GBP,-6
        """
        importer = csv.CSVImporter({Col.DATE: 'Posting Date',
                                    Col.NARRATION1: 'Description',
                                    Col.AMOUNT: 'Amount',
                                    Col.CURRENCY: 'Currency',
                                    Col.BALANCE: 'Balance'},
                                   'Assets:Bank',
                                   'PLN')
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

           2016-03-18 * "1st Payment in GBP"
             Assets:Bank  -1.00 GBP

           2016-03-18 * "1st Payment in PLN"
             Assets:Bank  -1 PLN

           2016-03-18 * "1st Payment in ZAR"
             Assets:Bank  -1.0 ZAR

           2016-03-19 * "2nd Payment in GBP"
             Assets:Bank  -2 GBP

           2016-03-19 * "2nd Payment in Main Currency"
             Assets:Bank  -2.00 PLN

           2016-03-20 * "3rd Payment in GBP"
             Assets:Bank  -3 GBP

           2016-03-20 balance Assets:Bank  -3 PLN

           2016-03-19 balance Assets:Bank  -1 ZAR

           2016-03-21 balance Assets:Bank  -6 GBP

        """, entries)

    @test_utils.docfile
    def test_zero_balance_assertion_is_added_with_currency_field(self, filename):
        """\
          Posting Date,"Description",Amount,Currency,Balance
          3/18/2016,"1st Payment in GBP",-1.00,GBP,-1
          3/18/2016,"1st Payment in PLN",-1,PLN,-1
          3/18/2016,"1st Payment in ZAR",-1.0,ZAR,-1
          3/19/2016,"2nd Payment in GBP",-2,GBP,-3
          3/19/2016,"2nd Payment in Main Currency",-2.00,,-3
          3/20/2016,"3rd Payment in GBP",-3,GBP,-6
          3/21/2016,"4th Payment in GBP",6,GBP,0
        """
        importer = csv.CSVImporter({Col.DATE: 'Posting Date',
                                    Col.NARRATION1: 'Description',
                                    Col.AMOUNT: 'Amount',
                                    Col.CURRENCY: 'Currency',
                                    Col.BALANCE: 'Balance'},
                                   'Assets:Bank',
                                   'PLN')
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

           2016-03-18 * "1st Payment in GBP"
             Assets:Bank  -1.00 GBP

           2016-03-18 * "1st Payment in PLN"
             Assets:Bank  -1 PLN

           2016-03-18 * "1st Payment in ZAR"
             Assets:Bank  -1.0 ZAR

           2016-03-19 * "2nd Payment in GBP"
             Assets:Bank  -2 GBP

           2016-03-19 * "2nd Payment in Main Currency"
             Assets:Bank  -2.00 PLN

           2016-03-20 * "3rd Payment in GBP"
             Assets:Bank  -3 GBP

           2016-03-21 * "4th Payment in GBP"
             Assets:Bank  6 GBP

           2016-03-20 balance Assets:Bank  -3 PLN

           2016-03-19 balance Assets:Bank  -1 ZAR

           2016-03-22 balance Assets:Bank  0 GBP

        """, entries)


    @test_utils.docfile
    def test_currency_and_balances_when_none_are_in_the_main_currency(self, filename):
        """\
          Posting Date,"Description",Amount,Currency,Balance
          3/18/2016,"1st Payment in GBP",-1.00,GBP,-1
          3/18/2016,"1st Payment in ZAR",-1.0,ZAR,-1
          3/19/2016,"2nd Payment in GBP",-2,GBP,-3
          3/20/2016,"3rd Payment in GBP",-3,GBP,-6
        """
        importer = csv.CSVImporter({Col.DATE: 'Posting Date',
                                    Col.NARRATION1: 'Description',
                                    Col.AMOUNT: 'Amount',
                                    Col.CURRENCY: 'Currency',
                                    Col.BALANCE: 'Balance'},
                                   'Assets:Bank',
                                   'PLN')
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

           2016-03-18 * "1st Payment in GBP"
             Assets:Bank  -1.00 GBP

           2016-03-18 * "1st Payment in ZAR"
             Assets:Bank  -1.0 ZAR

           2016-03-19 * "2nd Payment in GBP"
             Assets:Bank  -2 GBP

           2016-03-20 * "3rd Payment in GBP"
             Assets:Bank  -3 GBP

           2016-03-19 balance Assets:Bank  -1 ZAR

           2016-03-21 balance Assets:Bank  -6 GBP

        """, entries)

    @test_utils.docfile
    def test_main_currency_should_be_used_when_no_currency_is_specified(self, filename):
        """\
          Posting Date,"Description",Amount,Currency,Balance
          3/18/2016,"1st Payment",-1.00,,-1
          3/18/2016,"2nd Payment",-1.0,,-2
          3/19/2016,"3rd Payment",-2,,-4
          3/20/2016,"4th Payment",-3,,-7
        """
        importer = csv.CSVImporter({Col.DATE: 'Posting Date',
                                    Col.NARRATION1: 'Description',
                                    Col.AMOUNT: 'Amount',
                                    Col.CURRENCY: 'Currency',
                                    Col.BALANCE: 'Balance'},
                                   'Assets:Bank',
                                   'PLN')
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

           2016-03-18 * "1st Payment"
             Assets:Bank  -1.00 PLN

           2016-03-18 * "2nd Payment"
             Assets:Bank  -1.0 PLN

           2016-03-19 * "3rd Payment"
             Assets:Bank  -2 PLN

           2016-03-20 * "4th Payment"
             Assets:Bank  -3 PLN

           2016-03-21 balance Assets:Bank  -7 PLN

        """, entries)

    @test_utils.docfile
    def test_categorizer_one_argument(self, filename):
        """\
          Date,Amount,Payee,Description
          6/2/2020,30.00,"Payee here","Description"
          7/2/2020,-25.00,"Supermarket","Groceries"
        """
        def categorizer(txn):
            if txn.narration == "Groceries":
                txn.postings.append(
                    data.Posting("Expenses:Groceries",
                                 -txn.postings[0].units,
                                 None, None, None, None))

            return txn

        importer = csv.CSVImporter({Col.DATE: 'Date',
                                    Col.NARRATION: 'Description',
                                    Col.AMOUNT: 'Amount'},
                                   'Assets:Bank',
                                   'EUR',
                                   ('Date,Amount,Payee,Description'),
                                   categorizer=categorizer,
                                   institution='foobar')
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

          2020-06-02 * "Description"
            Assets:Bank  30.00 EUR

          2020-07-02 * "Groceries"
            Assets:Bank  -25.00 EUR
            Expenses:Groceries  25.00 EUR
        """, entries)

    @test_utils.docfile
    def test_categorizer_two_arguments(self, filename):
        """\
          Date,Amount,Payee,Description
          6/2/2020,30.00,"Payee here","Description"
          7/2/2020,-25.00,"Supermarket","Groceries"
        """
        def categorizer(txn, row):
            txn = txn._replace(payee=row[2])
            txn.meta['source'] = pformat(row)
            return txn

        importer = csv.CSVImporter({Col.DATE: 'Date',
                                    Col.NARRATION: 'Description',
                                    Col.AMOUNT: 'Amount'},
                                   'Assets:Bank',
                                   'EUR',
                                   ('Date,Amount,Payee,Description'),
                                   categorizer=categorizer,
                                   institution='foobar')
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

          2020-06-02 * "Payee here" "Description"
            source: "['6/2/2020', '30.00', 'Supermarket', 'Groceries']"
            Assets:Bank  30.00 EUR

          2020-07-02 * "Supermarket" "Groceries"
            source: "['7/2/2020', '-25.00', 'Supermarket', 'Groceries']"
            Assets:Bank  -25.00 EUR
        """, entries)

    @test_utils.docfile
    def test_explict_encoding_utf8(self, filename):
        """\
          Posting,Description,Amount
          2020/08/08,🍏,2
        """
        importer = csv.CSVImporter({Col.DATE: 'Posting',
                                    Col.NARRATION: 'Description',
                                    Col.AMOUNT: 'Amount'},
                                   'Assets:Bank', 'EUR', [],
                                   encoding='utf-8')
        entries = importer.extract(filename)
        self.assertEqualEntries(r"""

          2020-08-08 * "🍏"
            Assets:Bank  2 EUR

        """, entries)

    def test_newlines(self):
        content = textwrap.dedent("""\
          Date,Description,Amount
          2020-07-03,A,2
          2020-07-03,B,3
        """)
        importer = csv.CSVImporter({Col.DATE: 'Date',
                                    Col.NARRATION: 'Description',
                                    Col.AMOUNT: 'Amount'},
                                   'Assets:Bank', 'EUR', [])
        for nl in '\n', '\r\n', '\r':
            with tempfile.NamedTemporaryFile('w') as temp:
                temp.write(content.replace('\n', nl))
                temp.flush()
                entries = importer.extract(temp.name)
                self.assertEqualEntries("""
                  2020-07-03 * "A"
                    Assets:Bank  2 EUR

                  2020-07-03 * "B"
                    Assets:Bank  3 EUR
                """, entries)


# TODO: Test things out with/without payee and with/without narration.
# TODO: Test balance support.
# TODO: Add balances every month or week.
# TODO: Test ascending and descending orders.
# TODO: Add a test for all the supported fields, e.g. NARRATION2.



#!/usr/bin/env python3
"""Migrate old test files to the new one.

This script can be used to convert old test files for beancount.ingest to the
newer scheme where a single expected file is used. For example, this converts:

    2015-06-13.ofx.qbo
    2015-06-13.ofx.qbo.extract
    2015-06-13.ofx.qbo.file_account
    2015-06-13.ofx.qbo.file_date
    2015-06-13.ofx.qbo.file_name

to:

    2015-06-13.ofx.qbo
    2015-06-13.ofx.qbo.beancount

"""

import functools
import os
import re
from os import path
from typing import List

import click


def read_or_empty(filename: str) -> str:
    if path.exists(filename):
        with open(filename) as infile:
            return infile.read().rstrip()
    else:
        return ''


def process_files(filename: str):
    """Rename files around the given absolute filename."""
    account = read_or_empty(filename + ".file_account")
    date = read_or_empty(filename + ".file_date")
    name = read_or_empty(filename + ".file_name")
    extract = read_or_empty(filename + ".extract")
    with open(filename + ".beancount", "w") as outfile:
        pr = functools.partial(print, file=outfile)
        pr(';; Account: {}'.format(account or''))
        pr(';; Date: {}'.format(date or ''))
        pr(';; Name: {}'.format(name or ''))
        if extract:
            pr(extract)
    for ext in [".file_account", ".file_date", ".file_name", ".extract"]:
        if path.exists(filename + ext):
            os.remove(filename + ext)


@click.command()
@click.argument('directories',type=click.Path(exists=True, resolve_path=True), nargs=-1)
def main(directories: List[str]):
    for directory in directories:
        for root, dirs, files in os.walk(directory):
            for filename in sorted(files):
                afilename = path.join(root, filename)
                if re.search(r"\.(py|beancount)$", afilename):
                    continue
                if re.search(r"\.(extract|file_account|file_name|file_date)", afilename):
                    continue
                if path.exists(afilename + ".beancount"):
                    continue
                process_files(afilename)


if __name__ == '__main__':
    main()



# TODO(blais): If we don't remove this file it should get renamed to archive.py.
"""Base class that implements filing account.

It also sports an optional prefix to prepend to the renamed filename. Typically
you can put the name of the institution there, so you get a renamed filename
like this:

  YYYY-MM-DD.institution.Original_File_Name.pdf

"""
__author__ = "Martin Blais <blais@furius.ca>"

from os import path

from beancount.core import account
from beangulp import importer


class FilingMixin(importer.ImporterProtocol):

    def __init__(self, **kwds):
        """Pull 'filing' and 'prefix' from kwds.

        Args:
          filing: The name of the account to file to.
          prefix: The name of the institution prefix to insert.
        """

        self.filing_account = kwds.pop('filing', None)
        assert account.is_valid(self.filing_account)

        self.prefix = kwds.pop('prefix', None)

        super().__init__(**kwds)

    def name(self):
        """Include the filing account in the name."""
        return '{}: "{}"'.format(super().name(), self.filing_account)

    def file_account(self, file):
        return self.filing_account

    def file_name(self, file):
        """Return the optional renamed account filename."""
        supername = super().file_name(file)
        if not self.prefix:
            return supername
        return '.'.join(filter(None, [self.prefix,
                                      supername or path.basename(file.name)]))



import datetime
import decimal
import unittest

from beancount.core import data
from beancount.parser import cmptest
from beancount.utils.test_utils import docfile
from beangulp.importers.csvbase import Column, Columns, Date, Amount, CSVMeta, CSVReader, Order, Importer


class TestColumn(unittest.TestCase):

    def test_index_spec(self):
        column = Column(0)
        func = column.getter(None)
        value = func(('0', '1', '2', '3', ))
        self.assertEqual(value, '0')

    def test_named_spec(self):
        column = Column('Num')
        func = column.getter({'Num': 1})
        value = func(('0', '1', '2', '3', ))
        self.assertEqual(value, '1')

    def test_named_errrors(self):
        column = Column('A')
        with self.assertRaisesRegex(KeyError, "Cannot find column 'A' in "):
            column.getter({'B': 0, 'C': 1})
        with self.assertRaisesRegex(KeyError, "Column 'A' cannot be found in "):
            column.getter(None)

    def test_strip(self):
        # The default field parser strips whitespace.
        column = Column(0)
        func = column.getter(None)
        value = func((' value ', ))
        self.assertEqual(value, 'value')

    def test_default(self):
        column = Column(0, default=42)
        func = column.getter(None)
        value = func(('', ))
        self.assertEqual(value, 42)


class TestDateColumn(unittest.TestCase):

    def test_default_format(self):
        column = Date(0)
        func = column.getter(None)
        value = func(('2021-05-16', ))
        self.assertEqual(value, datetime.date(2021, 5, 16))

    def test_custom_format(self):
        column = Date(0, '%d.%m.%Y')
        func = column.getter(None)
        value = func(('16.05.2021', ))
        self.assertEqual(value, datetime.date(2021, 5, 16))


class TestColumnsColumn(unittest.TestCase):

    def test_default_sep(self):
        column = Columns(0, 1)
        func = column.getter(None)
        value = func(('0', '1', '2', '3', ))
        self.assertEqual(value, '0 1')

    def test_custom_sep(self):
        column = Columns(0, 1, sep=': ')
        func = column.getter(None)
        value = func(('0', '1', '2', '3', ))
        self.assertEqual(value, '0: 1')


class TestAmountColumn(unittest.TestCase):

    def test_parse_decimal(self):
        column = Amount(0)
        func = column.getter(None)
        value = func(('1.0', ))
        self.assertIsInstance(value, decimal.Decimal)
        self.assertEqual(value, decimal.Decimal('1.0'))

    def test_parse_subs_one(self):
        column = Amount(0, subs={',': ''})
        func = column.getter(None)
        value = func(('1,000.00', ))
        self.assertIsInstance(value, decimal.Decimal)
        self.assertEqual(value, decimal.Decimal('1000.00'))

    def test_parse_subs_two(self):
        column = Amount(0, subs={'\\.': '', ',': '.'})
        func = column.getter(None)
        value = func(('1.000,00', ))
        self.assertIsInstance(value, decimal.Decimal)
        self.assertEqual(value, decimal.Decimal('1000.00'))

    def test_parse_subs_currency(self):
        column = Amount(0, subs={'\\$(.*)': '\\1', ',': ''})
        func = column.getter(None)
        value = func(('$1,000.00', ))
        self.assertIsInstance(value, decimal.Decimal)
        self.assertEqual(value, decimal.Decimal('1000.00'))


class TestCSVMeta(unittest.TestCase):

    def test_collect_fields(self):
        class CSVTest(metaclass=CSVMeta):
            first = Column(0)
            second = Column(1)
        # pylint: disable=no-member
        self.assertEqual(CSVTest.columns.keys(), {'first', 'second'})


class TestCSVReader(unittest.TestCase):

    @docfile
    def test_named_columns(self, filename):
        """\
        First, Second
        1, 2
        3, 4
        """
        class Reader(CSVReader):
            first = Column('First')
            second = Column('Second')
        reader = Reader()
        rows = list(reader.read(filename))
        self.assertEqual(len(rows), 2)
        # tuple elements
        self.assertEqual(rows[0][0], '1')
        # leading space is preserved
        self.assertEqual(rows[1][1], ' 4')
        # attribute accessors
        self.assertEqual(rows[0].first, '1')
        self.assertEqual(rows[1].second, '4')

    @docfile
    def test_named_no_enough_lines(self, filename):
        """\
        # comment
        """
        class Reader(CSVReader):
            first = Column('First')
            second = Column('Second')
        reader = Reader()
        with self.assertRaisesRegex(IndexError, 'The input file does not contain '):
            list(reader.read(filename))

    @docfile
    def test_indexed_columns_names_false(self, filename):
        """\
        First, Second
        1, 2
        3, 4
        """
        class Reader(CSVReader):
            first = Column(0)
            second = Column(1)
            names = False
        reader = Reader()
        rows = list(reader.read(filename))
        self.assertEqual(len(rows), 3)
        # tuple elements
        self.assertEqual(rows[0][0], 'First')
        # leading space is preserved
        self.assertEqual(rows[2][1], ' 4')
        # attribute accessors
        self.assertEqual(rows[0].first, 'First')
        self.assertEqual(rows[2].second, '4')

    @docfile
    def test_indexed_columns_names_true(self, filename):
        """\
        First, Second
        1, 2
        3, 4
        """
        class Reader(CSVReader):
            first = Column(0)
            second = Column(1)
        reader = Reader()
        rows = list(reader.read(filename))
        self.assertEqual(len(rows), 2)
        # tuple elements
        self.assertEqual(rows[0][0], '1')
        # leading space is preserved
        self.assertEqual(rows[1][1], ' 4')
        # attribute accessors
        self.assertEqual(rows[0].first, '1')
        self.assertEqual(rows[1].second, '4')

    @docfile
    def test_comments(self, filename):
        """\
        First, Second
        # ignore
        a, b
        c, d
        """
        class Reader(CSVReader):
            first = Column(0)
            second = Column(1)
        reader = Reader()
        rows = list(reader.read(filename))
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0][0], 'a')

    @docfile
    def test_no_comments(self, filename):
        """\
        First, Second
        # ignore
        a, b
        c, d
        """
        class Reader(CSVReader):
            first = Column(0)
            second = Column(1)
            comments = False
        reader = Reader()
        rows = list(reader.read(filename))
        self.assertEqual(len(rows), 3)
        self.assertEqual(rows[0][0], '# ignore')

    @docfile
    def test_custom_comments(self, filename):
        """\
        First, Second
        # ignore
        ; ignore
        a, b
        c, d
        """
        class Reader(CSVReader):
            first = Column(0)
            second = Column(1)
            comments = ';'
        reader = Reader()
        rows = list(reader.read(filename))
        self.assertEqual(len(rows), 3)
        self.assertEqual(rows[0][0], '# ignore')

    @docfile
    def test_skiplines(self, filename):
        """\
        Skip
        Skip
        First, Second
        a, b
        """
        class Reader(CSVReader):
            first = Column(0)
            second = Column(1)
            skiplines = 2
        reader = Reader()
        rows = list(reader.read(filename))
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0][0], 'a')


class Base(Importer):
    def identify(self, filepath):
        return True


class TestImporter(cmptest.TestCase):

    @docfile
    def test_date(self, filename):
        """\
        2021-05-17, Test, 1.00
        2021-05-18, Test, 1.00
        2021-05-16, Test, 1.00
        """
        class CSVImporter(Base):
            date = Date(0)
        importer = CSVImporter('Account:CSV', 'EUR')
        date = importer.date(filename)
        self.assertEqual(date, datetime.date(2021, 5, 18))

    @docfile
    def test_extract(self, filename):
        """\
        2021-05-17, Test, 1.00
        """
        class CSVImporter(Base):
            date = Date(0)
            narration = Column(1)
            amount = Amount(2)
            names = False
        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqualEntries(entries, """
        2021-05-17 * "Test"
          Assets:CSV  1.00 EUR
        """)

    @docfile
    def test_extract_payee(self, filename):
        """\
        2021-05-17, Payee, Test, 1.00
        """
        class CSVImporter(Base):
            date = Date(0)
            payee = Column(1)
            narration = Column(2)
            amount = Amount(3)
            names = False
        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqualEntries(entries, """
        2021-05-17 * "Payee" "Test"
          Assets:CSV  1.00 EUR
        """)

    @docfile
    def test_extract_account(self, filename):
        """\
        2021-05-17, Test, Assets:Test, 1.00
        """
        class CSVImporter(Base):
            date = Date(0)
            narration = Column(1)
            account = Column(2)
            amount = Amount(3)
            names = False
        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqualEntries(entries, """
        2021-05-17 * "Test"
          Assets:Test  1.00 EUR
        """)

    @docfile
    def test_extract_balance(self, filename):
        """\
        2021-05-18, Test A, 2.00, 3.00
        2021-05-17, Test B, 1.00, 1.00
        """
        class CSVImporter(Base):
            date = Date(0)
            narration = Column(1)
            amount = Amount(2)
            balance = Amount(3)
            names = False
        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqualEntries(entries, """
        2021-05-17 * "Test B"
          Assets:CSV  1.00 EUR
        2021-05-18 * "Test A"
          Assets:CSV  2.00 EUR
        2021-05-19 balance Assets:CSV  3.00 EUR
        """)

    @docfile
    def test_extract_balance_with_multiple_transactions_per_day_and_ascending_dates(self, filename):
        """\
        2021-05-17, Test A, 1.00, 1.00
        2021-05-18, Test B, 2.00, 3.00
        2021-05-18, Test C, 1.00, 4.00
        """
        class CSVImporter(Base):
            date = Date(0)
            narration = Column(1)
            amount = Amount(2)
            balance = Amount(3)
            names = False
        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqualEntries(entries, """
        2021-05-17 * "Test A"
          Assets:CSV  1.00 EUR
        2021-05-18 * "Test B"
          Assets:CSV  2.00 EUR
        2021-05-18 * "Test C"
          Assets:CSV  1.00 EUR
        2021-05-19 balance Assets:CSV  4.00 EUR
        """)

    @docfile
    def test_extract_balance_with_order_override(self, filename):
        """\
        2021-05-18, Test B, 2.00, 4.00
        2021-05-18, Test C, 1.00, 2.00
        """
        class CSVImporter(Base):
            date = Date(0)
            narration = Column(1)
            amount = Amount(2)
            balance = Amount(3)
            names = False
            order = Order.DESCENDING
        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqualEntries(entries, """
        2021-05-18 * "Test C"
          Assets:CSV  1.00 EUR
        2021-05-18 * "Test B"
          Assets:CSV  2.00 EUR
        2021-05-19 balance Assets:CSV  4.00 EUR
        """)

    @docfile
    def test_extract_flag(self, filename):
        """\
        2021-05-17, Test, 1.00, !
        """
        class CSVImporter(Base):
            date = Date(0)
            narration = Column(1)
            amount = Amount(2)
            flag = Column(3)
            names = False
        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqualEntries(entries, """
        2021-05-17 ! "Test"
          Assets:CSV  1.00 EUR
        """)

    @docfile
    def test_extract_link_and_tag(self, filename):
        """\
        2021-05-17, Test, 1.00, link, tag
        """
        class CSVImporter(Base):
            date = Date(0)
            narration = Column(1)
            amount = Amount(2)
            link = Column(3)
            tag = Column(4)
            names = False
        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqualEntries(entries, """
        2021-05-17 * "Test" ^link #tag
          Assets:CSV  1.00 EUR
        """)

    @docfile
    def test_extract_currency(self, filename):
        """\
        2021-05-17, Test A US, 1.00, 1.00, USD
        2021-05-17, Test A EU, 2.00, 2.00, EUR
        2021-05-18, Test B US, 2.00, 3.00, USD
        2021-05-18, Test B EU, 3.00, 5.00, EUR
        """
        class CSVImporter(Base):
            date = Date(0)
            narration = Column(1)
            amount = Amount(2)
            balance = Amount(3)
            currency = Column(4)
            names = False
        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqualEntries(entries, """
        2021-05-17 * "Test A US"
          Assets:CSV  1.00 USD
        2021-05-17 * "Test A EU"
          Assets:CSV  2.00 EUR
        2021-05-18 * "Test B US"
          Assets:CSV  2.00 USD
        2021-05-18 * "Test B EU"
          Assets:CSV  3.00 EUR
        2021-05-19 balance Assets:CSV  3.00 USD
        2021-05-19 balance Assets:CSV  5.00 EUR
        """)

    @docfile
    def test_extract_metadata(self, filename):
        """\
        2021-05-17, Test, 1.00, data, 42
        """
        class CSVImporter(Base):
            date = Date(0)
            narration = Column(1)
            amount = Amount(2)
            meta = Column(3)
            data = Amount(4)
            names = False

            def metadata(self, filepath, lineno, row):
                meta = super().metadata(filepath, lineno, row)
                for field in 'meta', 'data':
                    meta[field] = getattr(row, field)
                return meta

        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqualEntries(entries, """
        2021-05-17 * "Test"
          meta: "data"
          data: 42
          Assets:CSV  1.00 EUR
        """)

    @docfile
    def test_extract_finalize(self, filename):
        """\
        2021-05-17, Test, -1.00, Testing
        """
        class CSVImporter(Base):
            date = Date(0)
            narration = Column(1)
            amount = Amount(2)
            category = Column(3)
            names = False

            def finalize(self, txn, row):
                posting = data.Posting(
                    'Expenses:' + row.category,
                    # This could be None in a real importer. However,
                    # the trsting framework accepts only complete
                    # transactions, thus we do the booking manually.
                    -txn.postings[0].units,
                    None, None, None, None)
                txn.postings.append(posting)
                return txn

        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqualEntries(entries, """
        2021-05-17 * "Test"
          Assets:CSV  -1.00 EUR
          Expenses:Testing  1.00 EUR
        """)

    @docfile
    def test_extract_finalize_remove(self, filename):
        """\
        2021-05-17, Test, -1.00, Testing
        2021-05-18, Drop, -2.00, Testing
        """
        class CSVImporter(Base):
            date = Date(0)
            narration = Column(1)
            amount = Amount(2)
            names = False

            def finalize(self, txn, row):
                if txn.narration == 'Drop':
                    return None
                return txn

        importer = CSVImporter('Assets:CSV', 'EUR')
        entries = importer.extract(filename, [])
        self.assertEqual(len(entries), 1)



__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import unittest

from beangulp import importer
from beangulp import cache


class TestImporterProtocol(unittest.TestCase):

    def test_importer_methods(self):
        # Kind of a dumb test, but for consistency we just test everything.
        memo = cache._FileMemo('/tmp/test')
        imp = importer.ImporterProtocol()
        self.assertIsInstance(imp.FLAG, str)
        self.assertFalse(imp.identify(memo))
        self.assertFalse(imp.extract(memo))
        self.assertFalse(imp.file_account(memo))
        self.assertFalse(imp.file_date(memo))
        self.assertFalse(imp.file_name(memo))



"""Identify similar entries.

This can be used during import in order to identify and flag duplicate entries.
"""

__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

from decimal import Decimal
from typing import Callable, Optional
import collections
import datetime
import re

from beancount.core.number import ZERO
from beancount.core.number import ONE
from beancount.core import data
from beancount.core import amount
from beancount.core import interpolate


try:
    from functools import cache
except ImportError:
    from functools import lru_cache

    cache = lru_cache(maxsize=None)


def find_similar_entries(entries, existing_entries, cmp=None, window_days=2):
    """Find which entries from a list are potential duplicates of a set.

    The existing_entries array must be sorted by date. If there are
    multiple entries in existing_entries matching an entry in entries,
    only the first match is returned.

    Args:
      entries: The list of entries to classify as duplicate or note.
      existing_entries: The list of entries against which to match.
      cmp: A functor used to establish the similarity of two entries.
      window_days: The number of days (inclusive) before or after to scan the
        entries to classify against.

    Returns:
      A list of (entry, existing_entry) tuples where entry is from
      entries and is deemed to be a duplicate of existing_entry, from
      existing_entries.

    """
    window_head = datetime.timedelta(days=window_days)
    window_tail = datetime.timedelta(days=window_days + 1)

    if cmp is None:
        cmp = heuristic_comparator()

    # For each of the new entries, look at existing entries at a nearby date.
    duplicates = []
    if existing_entries is not None:
        for entry in data.filter_txns(entries):
            for existing_entry in data.filter_txns(
                data.iter_entry_dates(
                    existing_entries, entry.date - window_head, entry.date + window_tail
                )
            ):
                if cmp(entry, existing_entry):
                    duplicates.append((entry, existing_entry))
                    break
    return duplicates


class hashable:
    def __init__(self, obj):
        self.obj = obj

    def __hash__(self):
        return id(self.obj)

    def __getattr__(self, name):
        return getattr(self.obj, name)


Comparator = Callable[[data.Directive, data.Directive], bool]


def heuristic_comparator(
    max_date_delta: Optional[datetime.timedelta] = None, epsilon: Optional[Decimal] = None
) -> Comparator:
    """Generic comparison function generator.

    Two transactions are deemed similar if

    - their dates are within a close range of each other (e.g. 2 days), if
      specified with `max_date_delta`,

    - amounts on postings corresponding to the same account are within some
      fraction of each other (default: 5%), and

    - the set of accounts of the two transactions are the same or one is a
      subset of the other.

    Args:
      max_date_delta: A timedelta datetime difference within which two
        transactions may be considered similar.
      epsilon: A Decimal fraction representing how close the amounts are
        required to be of each other. For example, Decimal("0.01") for 1%.
    Returns:
      A comparator predicate accepting two directives and returning a bool.
    """

    if epsilon is None:
        epsilon = Decimal("0.05")

    def cmp(entry1: data.Directive, entry2: data.Directive) -> bool:
        """Compare two entries.

        Implement a heuristic method to determine if two transactions are
        similar enough to be considered duplicates.
        """
        # This comparator needs to be able to handle Transaction
        # instances which are incomplete on one side, which have
        # slightly different dates, or potentially postings with
        # slightly different amounts.

        if not isinstance(entry1, data.Transaction) or not isinstance(
            entry2, data.Transaction
        ):
            return False

        # Check the date difference.
        if max_date_delta is not None:
            delta = (
                (entry1.date - entry2.date)
                if entry1.date > entry2.date
                else (entry2.date - entry1.date)
            )
            if delta > max_date_delta:
                return False

        amounts1 = amounts_map_cached(hashable(entry1))
        amounts2 = amounts_map_cached(hashable(entry2))

        # Look for amounts on common accounts.
        common_keys = set(amounts1) & set(amounts2)
        for key in sorted(common_keys):
            # Compare the amounts.
            number1 = amounts1[key]
            number2 = amounts2[key]
            if number1 == ZERO and number2 == ZERO:
                break
            diff = abs((number1 / number2) if number2 != ZERO else (number2 / number1))
            if diff == ZERO:
                return False
            if diff < ONE:
                diff = ONE / diff
            if (diff - ONE) < epsilon:
                break
        else:
            return False

        # Here, we have found at least one common account with a close
        # amount. Now, we require that the set of accounts are equal or that
        # one be a subset of the other.
        accounts1 = {posting.account for posting in entry1.postings}
        accounts2 = {posting.account for posting in entry2.postings}
        return accounts1.issubset(accounts2) or accounts2.issubset(accounts1)

    return cmp


# Old alias to the heuristic comparator kept for backwards compatibility.
comparator = heuristic_comparator


def same_link_comparator(regex: Optional[str] = None) -> Comparator:
    """Comparison function generator that checks if two directives share a link.

    You can use this if you have a source of transactions that consistently
    defined and unique transaction which it produces as a link. The matching of
    these transactions will be more precise than the heuristic no dates and
    amounts used by default, if you keep the links in your ledger.

    You can further restrict the set of links that are compared if you provide a
    regex. This can be useful if your importer produces multiple other links.

    Args:
      regex: An optional regular expression used to filter the links.
    Returns:
      A comparator predicate accepting two directives and returning a bool.
    """

    def cmp(entry1: data.Directive, entry2: data.Directive) -> bool:
        """Compare two entries by common link."""

        if not isinstance(entry1, data.Transaction) or not isinstance(
            entry2, data.Transaction
        ):
            return False

        links1 = entry1.links
        links2 = entry2.links
        if regex:
            links1 = {link for link in links1 if re.match(regex, link)}
            links2 = {link for link in links2 if re.match(regex, link)}

        return bool(links1 & links2)

    return cmp


def amounts_map(entry):
    """Compute a mapping of (account, currency) -> Decimal balances.

    Args:
      entry: A Transaction instance.
    Returns:
      A dict of account -> Amount balance.
    """
    amounts = collections.defaultdict(Decimal)
    for posting in entry.postings:
        # Skip interpolated postings.
        if posting.meta and interpolate.AUTOMATIC_META in posting.meta:
            continue
        currency = isinstance(posting.units, amount.Amount) and posting.units.currency
        if isinstance(currency, str):
            key = (posting.account, currency)
            amounts[key] += posting.units.number
    return amounts


amounts_map_cached = cache(amounts_map)



"""Test loader to make unittest test discovery pick up the doctests."""

import doctest
import unittest


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    suite.addTest(
        doctest.DocFileSuite(
            'archive.rst',
            'extract.rst',
            'identify.rst',
            'testing.rst',
            optionflags=(
                doctest.ELLIPSIS |
                doctest.NORMALIZE_WHITESPACE |
                doctest.REPORT_NDIFF |
                # Display only the first failed test. Note that all
                # tests are run, thus the cleanup at the end of the
                # doctests file is still executed.
                doctest.REPORT_ONLY_FIRST_FAILURE)))
    return suite



"""Base class that implements identification using regular expressions.
"""
__author__ = "Martin Blais <blais@furius.ca>"

import collections
import itertools
import re

from beangulp import cache
from beangulp import importer


_PARTS = {'mime', 'filename', 'content'}


def identify(remap, converter, file):
    """Identify the contents of a file.

    Args:
      remap: A dict of 'part' to list-of-compiled-regexp objects, where each item is
        a specification to match against its part. The 'part' can be one of 'mime',
        'filename' or 'content'.
      converter: A function to convert the given file.
    Returns:
      A boolean, true if the file is not rejected by the constraints.
    """
    if remap.get('mime', None):
        mimetype = file.convert(cache.mimetype)
        if mimetype is None:
            return False
        if not all(regexp.search(mimetype) for regexp in remap['mime']):
            return False

    if remap.get('filename', None):
        if not all(regexp.search(file.name) for regexp in remap['filename']):
            return False

    if remap.get('content', None):
        # If this is a text file, read the whole thing in memory.
        text = file.convert(converter or cache.contents)
        if not all(regexp.search(text) for regexp in remap['content']):
            return False

    return True


class IdentifyMixin(importer.ImporterProtocol):

    def __init__(self, **kwds):
        """Pull 'matchers' and 'converter' from kwds."""

        self.remap = collections.defaultdict(list)
        matchers = kwds.pop('matchers', [])
        cls_matchers = getattr(self, 'matchers', [])
        assert isinstance(matchers, list)
        assert isinstance(cls_matchers, list)
        for part, regexp in itertools.chain(matchers, cls_matchers):
            assert part in _PARTS, repr(part)
            assert isinstance(regexp, str), repr(regexp)
            self.remap[part].append(re.compile(regexp))

        # Converter is a fn(filename: Text) -> contents: Text.
        self.converter = kwds.pop('converter', getattr(self, 'converter', None))

        super().__init__(**kwds)

    def identify(self, file):
        return identify(self.remap, self.converter, file)



__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

from decimal import Decimal as D
import datetime
import re

from beancount.core import data
from beancount.parser import cmptest
from beancount.parser import parser
from beancount import loader

from beangulp import similar


class TestDups(cmptest.TestCase):

    @loader.load_doc()
    def test_find_similar_entries(self, entries, _, __):
        """
        plugin "beancount.plugins.auto_accounts"

        2016-01-03 *
          Expenses:Tips         1.03 USD
          Assets:Other

        2016-01-04 *
          Expenses:Coffee       1.04 USD
          Assets:Other

        2016-01-05 *
          Expenses:Restaurant   1.05 USD
          Assets:Other

        2016-01-06 *
          Expenses:Groceries    1.06 USD
          Assets:Other

        2016-01-07 *
          Expenses:Alcohol      1.07 USD
          Assets:Other

        2016-01-08 *
          Expenses:Smoking      1.08 USD
          Assets:Other

        2016-01-09 *
          Expenses:Taxi         1.09 USD
          Assets:Other
        """
        new_entries, _, __ = loader.load_string(
            """
            plugin "beancount.plugins.auto_accounts"

            2016-01-06 *
              Expenses:Groceries    1.06 USD
              Assets:Other
        """
        )
        for days, num_comparisons in [(0, 1), (1, 1), (2, 1)]:
            duplicates = similar.find_similar_entries(
                new_entries, entries, lambda e1, e2: True, window_days=days
            )
            self.assertEqual(num_comparisons, len(duplicates))

            duplicates = similar.find_similar_entries(
                new_entries, entries, lambda e1, e2: False, window_days=days
            )
            self.assertEqual(0, len(duplicates))

    @loader.load_doc()
    def test_find_similar_entries__multiple_matches(self, entries, _, __):
        """
        plugin "beancount.plugins.auto_accounts"

        2016-02-01 * "A"
          Assets:Account1    10.00 USD
          Assets:Account2   -10.00 USD

        2016-02-02 * "B"
          Assets:Account1    10.00 USD
          Assets:Account2   -10.00 USD

        2016-02-03 * "C"
          Assets:Account1    10.00 USD
          Assets:Account2   -10.00 USD

        2016-02-04 * "D"
          Assets:Account1    10.00 USD
          Assets:Account2   -10.00 USD

        2016-02-05 * "D"
          Assets:Account1    10.00 USD
          Assets:Account2   -10.00 USD
        """
        # Test it with a single entry.
        new_entries = list(data.filter_txns(entries))[2:3]
        duplicates = similar.find_similar_entries(new_entries, entries, window_days=1)
        self.assertEqual(1, len(duplicates))
        self.assertEqual(new_entries[0], duplicates[0][0])

        # Test it with multiple duplicate entries.
        new_entries = list(data.filter_txns(entries))[1:4]
        duplicates = similar.find_similar_entries(new_entries, entries, window_days=1)
        self.assertEqual(len(new_entries), len(duplicates))

    @parser.parse_doc(allow_incomplete=True)
    def test_amounts_map(self, entries, _, __):
        """
        plugin "beancount.plugins.auto_accounts"

        2016-01-03 *
          Expenses:Alcohol     20.00 USD
          Expenses:Tips         1.03 USD
          Assets:Other

        2016-01-03 *
          Expenses:Tips         1.01 USD
          Expenses:Tips         1.02 USD
          Assets:Other
        """
        txns = list(data.filter_txns(entries))
        amap = similar.amounts_map(txns[0])
        self.assertEqual(
            {
                ("Expenses:Tips", "USD"): D("1.03"),
                ("Expenses:Alcohol", "USD"): D("20.00"),
            },
            amap,
        )

        amap = similar.amounts_map(txns[1])
        self.assertEqual({("Expenses:Tips", "USD"): D("2.03")}, amap)


class TestHeuristicComparator(cmptest.TestCase):

    def setUp(self):
        self.comparator = similar.heuristic_comparator(datetime.timedelta(days=2))

    @loader.load_doc()
    def test_simple(self, entries, _, __):
        """
        plugin "beancount.plugins.auto_accounts"

        2016-01-03 * "Base reservation" ^base
          Expenses:Alcohol     20.00 USD
          Expenses:Tips         1.03 USD
          Assets:Other

        2016-01-03 * "Similar amount within bounds" ^in-bounds
          Expenses:Alcohol     20.99 USD
          Assets:Other
        2016-01-03 * "Similar amount out of bounds" ^out-bounds
          Expenses:Alcohol     21.00 USD
          Assets:Other

        2016-01-06 * "Date too far" ^too-late
          Expenses:Alcohol     20.00 USD
          Expenses:Tips         1.03 USD
          Assets:Other

        2016-01-03 * "Non-overlapping accounts" ^non-accounts
          Expenses:Alcohol     20.00 USD
          Expenses:Tips         1.03 USD
          Assets:SomethingElse
        """
        txns = list(data.filter_txns(entries))

        def compare(expected, link1, link2):
            self.assertEqual(
                expected,
                self.comparator(
                    next(txn for txn in txns if link1 in txn.links),
                    next(txn for txn in txns if link2 in txn.links),
                ),
            )

        compare(True, "base", "base")
        compare(True, "base", "in-bounds")
        compare(False, "base", "out-bounds")
        compare(False, "base", "too-late")
        compare(False, "base", "non-accounts")


class TestSamelinkComparator(cmptest.TestCase):

    def setUp(self):
        self.comparator = similar.same_link_comparator(regex=r"\d+$")

    @loader.load_doc()
    def test_simple(self, entries, _, __):
        """
        2016-01-01 open Expenses:Alcohol
        2016-01-01 open Expenses:Food
        2016-01-01 open Expenses:Tips
        2016-01-01 open Assets:Other

        2016-01-03 * "Base" ^43567284654849564
          Expenses:Alcohol     20.00 USD
          Expenses:Tips         1.03 USD
          Assets:Other

        2016-01-04 * "Matching" ^43567284654849564
          Expenses:Food     25.00 USD
          Assets:Other

        2016-02-01 * "Matching with far date" ^43567284654849564
          Expenses:Food     25.00 USD
          Assets:Other

        2016-01-04 * "No links"
          Expenses:Food     25.00 USD
          Assets:Other

        2016-01-04 * "Not the same link" ^too-late
          Expenses:Food     25.00 USD
          Assets:Other

        2016-01-04 * "Difference number, matches regex" ^93580348349894834
          Expenses:Food     25.00 USD
          Assets:Other

        """
        txns = list(data.filter_txns(entries))
        for txn in filter(lambda t: re.match("Matching", t.narration), txns):
            self.assertEqual(True, self.comparator(txns[0], txn))
        for txn in filter(lambda t: not re.match("Matching", t.narration), txns[1:]):
            self.assertEqual(False, self.comparator(txns[0], txn))



from decimal import Decimal
from os import path
import datetime
import logging
import os
import types
import unittest
from unittest import mock
from shutil import rmtree
from tempfile import mkdtemp

from beangulp import utils


class TestWalk(unittest.TestCase):
    def setUp(self):
        self.temp = mkdtemp()

    def tearDown(self):
        rmtree(self.temp)

    def test_walk_empty(self):
        entries = utils.walk([self.temp])
        self.assertListEqual(list(entries), [])

    def test_walk_simple(self):
        filenames = [path.join(self.temp, name) for name in ('z', 'a', 'b')]
        for filename in filenames:
            with open(filename, 'w'):
                pass

        entries = utils.walk([self.temp])
        self.assertListEqual(list(entries), sorted(filenames))

        entries = utils.walk(filenames)
        self.assertListEqual(list(entries), filenames)

    def test_walk_subdir(self):
        os.mkdir(path.join(self.temp, 'dir'))
        filenames = [path.join(self.temp, 'dir', name) for name in ('a', 'b')]
        for filename in filenames:
            with open(filename, 'w'):
                pass

        entries = utils.walk([self.temp])
        self.assertListEqual(list(entries), filenames)

    def test_walk_mixed(self):
        os.mkdir(path.join(self.temp, 'dir'))
        files = ['c', ('dir', 'a'), ('dir', 'b')]
        filenames = [path.join(self.temp, *p) for p in files]
        for filename in filenames:
            with open(filename, 'w'):
                pass

        entries = utils.walk([self.temp])
        self.assertListEqual(list(entries), filenames)


class TestUtils(unittest.TestCase):

    def test_getmdate(self):
        self.assertIsInstance(utils.getmdate(__file__), datetime.date)

    def test_logger(self):
        logger = utils.logger()
        self.assertIsInstance(logger, types.FunctionType)
        logger = utils.logger(logging.INFO, err=True)
        self.assertIsInstance(logger, types.FunctionType)

    def test_sha1sum(self):
        self.assertRegex(utils.sha1sum(__file__), '[a-f0-9]+')

    def test_is_mimetype(self):
        self.assertTrue(utils.is_mimetype(__file__, {'text/x-python'}))
        self.assertTrue(utils.is_mimetype(__file__, 'text/x-python'))

    def test_search(self):
        self.assertTrue(utils.search_file_regexp(__file__, 'def test_search', encoding='utf8'))
        self.assertFalse(utils.search_file_regexp(__file__, '^$', encoding='utf8'))

    def test_parse_amount(self):
        self.assertEqual(Decimal('-1045.67'), utils.parse_amount('(1,045.67)'))

    def test_validate(self):
        utils.validate_accounts(
            {'cash': 'Cash account', 'position': 'Cash account'},
            {'cash': 'Assets:US:Cash', 'position': 'Assets:Investment'})

        # Missing values.
        with self.assertRaises(ValueError):
            utils.validate_accounts(
                {'cash': 'Cash account', 'position': 'Cash account'},
                {'position': 'Assets:Investment'})

        # Unknown values.
        with self.assertRaises(ValueError):
            utils.validate_accounts(
                {'cash': 'Cash account'},
                {'cash': 'Assets:US:Cash', 'position': 'Assets:Investment'})

        # Invalid values.
        with self.assertRaises(ValueError):
            utils.validate_accounts(
                {'cash': 'Cash account'},
                {'cash': 42})

    def test_idify(self):
        self.assertEqual(
            "A_great_movie_for_us.mp4", utils.idify(" A great movie (for us) .mp4 ")
        )
        self.assertEqual("A____B.pdf", utils.idify("A____B_._pdf"))


class TestDefDictWithKey(unittest.TestCase):
    def test_defdict_with_key(self):
        factory = mock.MagicMock()
        testdict = utils.DefaultDictWithKey(factory)

        testdict["a"]
        testdict["b"]
        self.assertEqual(2, len(factory.mock_calls))
        self.assertEqual(("a",), factory.mock_calls[0][1])
        self.assertEqual(("b",), factory.mock_calls[1][1])



"""Importer protocol definition."""

__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import abc
import datetime
import inspect

from typing import Optional

from beancount.core import flags
from beancount.core import data
from beangulp import cache
from beangulp import utils
from beangulp import extract
from beangulp import similar


class Importer(abc.ABC):
    """Interface that all source importers need to comply with.

    The interface is defined as an abstract base class implementing
    base behavior for all importers. Importer implementations need to
    provide at least the identify() and account() methods.

    """

    @property
    def name(self) -> str:
        """Unique id for the importer.

        The name is used to identify the importer in the command line
        interface. Conventionally this is a dotted string containing
        the module and name of the class, however a specific format is
        not enforced.

        """
        return f"{self.__class__.__module__}.{self.__class__.__name__}"

    @abc.abstractmethod
    def identify(self, filepath: str) -> bool:
        """Return True if this importer matches the given file.

        Args:
          filepath: Filesystem path to the document to be matched.

        Returns:
          True if this importer can handle this file.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def account(self, filepath: str) -> data.Account:
        """Return the account associated with the given file.

        The account is used to determine the archival folder for the
        document. While the interface allows returning different
        accounts for different documents, normally the returned
        account is a just a function of the importer instance.

        Args:
          filepath: Filesystem path to the document being imported.

        Returns:
          An account name.

        """
        raise NotImplementedError

    def date(self, filepath: str) -> Optional[datetime.date]:
        """Return the archival date the given file.

        The date is used by the archive command to form the archival
        filename of the document. If this method returns None, the
        date corresponding to the file document modification time is
        used.

        Args:
          filepath: Filesystem path to the document being imported.

        Returns:
          A date object or None.

        """
        return None

    def filename(self, filepath: str) -> Optional[str]:
        """Return the archival filename for the given file.

        Tidy filenames or rename documents when archiving them. This
        method should return a valid filename or None. In the latter
        case, the file path basename is used unmodified.

        Args:
          filepath: Filesystem path to the document being imported.

        Returns:
          The document filename to use for archiving.

        """
        return None

    def extract(self, filepath: str, existing: data.Entries) -> data.Entries:
        """Extract transactions and other directives from a document.

        The existing entries list is loaded from the existing ledger
        file, if the user specified one on the command line. It can be
        used to supplement the information provided by the document
        being processed to drive the extraction. For example to derive
        the prior state of the inventory.

        Args:
          filepath: Filesystem path to the document being imported.
          existing: Entries loaded from the existing ledger.

        Returns:
          A list of imported directives extracted from the document.

        """
        return []

    cmp = staticmethod(similar.heuristic_comparator())

    def deduplicate(self, entries: data.Entries, existing: data.Entries) -> None:
        """Mark duplicates in extracted entries.

        The default implementation uses the cmp() method to compare
        each newly extracted entries to the existing entries. Only
        existing entries dated within a 5 days window around the date
        of the each existing entry (two days prior and two days past)
        are considered.

        Duplicated entries are marked setting the "__duplicate__"
        entry metadata field to the entry of which the entry is a
        duplicate.

        Args:
          entries: Entries extracted from the document being processed.
          existing: Entries loaded from the existing ledger.

        """
        window = datetime.timedelta(days=2)
        extract.mark_duplicate_entries(entries, existing, window, self.cmp)

    def sort(self, entries: data.Entries, reverse=False) -> None:
        """Sort the extracted directives.

        The sort is in-place and stable. The reverse flag can be set
        to sort in descending order. Importers can implement this
        method to have entries serialized to file in a specific
        order. The default implementation sorts the entries according
        to beancount.core.data.entry_sortkey().

        Args:
          entries: Entries list to sort.
          reverse: When True sort in descending order.

        Returns:
          None.

        """
        return entries.sort(key=data.entry_sortkey, reverse=reverse)


class ImporterProtocol:
    """Old importers interface, superseded by the Importer ABC.

    The main difference is that the methods of this class accept a
    cache._FileMemo instance instead than the filesystem path to the
    imported document.

    """

    # A flag to use on new transaction. Override this flag in derived classes if
    # you prefer to create your imported transactions with a different flag.
    FLAG = flags.FLAG_OKAY

    def name(self):
        """See Importer class name property."""
        return f"{self.__class__.__module__}.{self.__class__.__name__}"

    __str__ = name

    def identify(self, file) -> bool:
        """See Importer class identify() method."""

    def file_account(self, file) -> data.Account:
        """See Importer class account() method."""

    def file_date(self, file) -> Optional[datetime.date]:
        """See Importer class date() method."""

    def file_name(self, file) -> Optional[str]:
        """See Importer class filename() method."""

    def extract(self, file, existing_entries: data.Entries = None) -> data.Entries:
        """See Importer class extract() method."""


class Adapter(Importer):
    """Adapter from ImporterProtocol to Importer ABC interface."""

    def __init__(self, importer):
        assert isinstance(importer, ImporterProtocol)
        self.importer = importer

    @property
    def name(self):
        return self.importer.name()

    def identify(self, filepath):
        return self.importer.identify(cache.get_file(filepath))

    def account(self, filepath):
        return self.importer.file_account(cache.get_file(filepath))

    def date(self, filepath):
        return self.importer.file_date(cache.get_file(filepath))

    def filename(self, filepath):
        filename = self.importer.file_name(cache.get_file(filepath))
        # The current implementation of the archive command does not
        # modify the filename returned by the importer. This preserves
        # backward compatibility with the old implementation of the
        # command.
        return utils.idify(filename) if filename else None

    def extract(self, filepath, existing):
        p = inspect.signature(self.importer.extract).parameters
        if len(p) > 1:
            return self.importer.extract(cache.get_file(filepath), existing)
        return self.importer.extract(cache.get_file(filepath))



__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import os
import shutil
import tempfile
import time
import unittest

from unittest import mock

from beangulp import cache
from beangulp import utils


class TestFileMemo(unittest.TestCase):

    def test_cache(self):
        with tempfile.NamedTemporaryFile() as tmpfile:
            shutil.copy(__file__, tmpfile.name)
            wrap = cache._FileMemo(tmpfile.name)

            # Check attributes.
            self.assertEqual(tmpfile.name, wrap.name)

            # Check that caching works.
            converter = mock.MagicMock(return_value='abc')
            self.assertEqual('abc', wrap.convert(converter))
            self.assertEqual('abc', wrap.convert(converter))
            self.assertEqual('abc', wrap.convert(converter))
            self.assertEqual(1, converter.call_count)

    def test_cache_head_and_contents(self):
        with tempfile.NamedTemporaryFile(suffix='.py') as tmpfile:
            shutil.copy(__file__, tmpfile.name)
            wrap = cache._FileMemo(tmpfile.name)

            contents = wrap.convert(cache.contents)
            self.assertIsInstance(contents, str)
            self.assertGreater(len(contents), 128)

            contents2 = wrap.contents()
            self.assertEqual(contents, contents2)

            head = wrap.convert(cache.head(128))
            self.assertIsInstance(head, str)
            self.assertEqual(128, len(head))

            mimetype = wrap.convert(cache.mimetype)
            self.assertIn(mimetype, {'text/plain',
                                     'text/x-python',
                                     'text/x-script.python',
                                     'text/c++'})

    def test_cache_head_obeys_explict_utf8_encoding_avoids_chardet_exception(self):
        data = b'asciiHeader1,\xf0\x9f\x8d\x8fHeader1,asciiHeader2'
        with mock.patch('builtins.open', mock.mock_open(read_data=data)):
            string = cache._FileMemo('filepath').head(encoding='utf-8')
            self.assertEqual(string, data.decode('utf8'))

    def test_cache_head_encoding(self):
        data = b'asciiHeader1,\xf0\x9f\x8d\x8fHeader1,asciiHeader2'
        # The 15th bytes is in the middle of the unicode character.
        num_bytes = 15
        with mock.patch('builtins.open', mock.mock_open(read_data=data)):
            string = cache._FileMemo('filepath').head(num_bytes, encoding='utf-8')
            self.assertEqual(string, 'asciiHeader1,')


class CacheTestCase(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        cache.CACHEDIR = os.path.join(self.tempdir, 'cache')
        os.mkdir(cache.CACHEDIR)
        self.filename = os.path.join(self.tempdir, 'test.txt')
        with open(self.filename, 'w'):
            pass

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_no_cache(self):
        counter = 0

        def func(filename):
            nonlocal counter
            counter +=1
            return counter

        r = func(self.filename)
        self.assertEqual(r, 1)

        r = func(self.filename)
        self.assertEqual(r, 2)

    def test_cache(self):
        counter = 0

        @cache.cache
        def func(filename):
            nonlocal counter
            counter +=1
            return counter

        r = func(self.filename)
        self.assertEqual(r, 1)

        r = func(self.filename)
        self.assertEqual(r, 1)

    def test_cache_expire_mtime(self):
        counter = 0

        @cache.cache
        def func(filename):
            nonlocal counter
            counter +=1
            return counter

        r = func(self.filename)
        self.assertEqual(r, 1)

        r = func(self.filename)
        self.assertEqual(r, 1)

        t = time.time() + 2.0
        os.utime(self.filename, (t, t))

        r = func(self.filename)
        self.assertEqual(r, 2)

        r = func(self.filename)
        self.assertEqual(r, 2)

    def test_cache_expire_args(self):
        counter = 0

        @cache.cache
        def func(filename, arg):
            nonlocal counter
            counter +=1
            return counter

        r = func(self.filename, 1)
        self.assertEqual(r, 1)

        r = func(self.filename, 1)
        self.assertEqual(r, 1)

        r = func(self.filename, 2)
        self.assertEqual(r, 2)

    def test_cache_expire_override(self):
        counter = 0

        @cache.cache
        def func(filename):
            nonlocal counter
            counter +=1
            return counter

        r = func(self.filename)
        self.assertEqual(r, 1)

        r = func(self.filename)
        self.assertEqual(r, 1)

        r = func(self.filename, cache=False)
        self.assertEqual(r, 2)

        r = func(self.filename, cache=True)
        self.assertEqual(r, 2)

        t = time.time() + 2.0
        os.utime(self.filename, (t, t))

        r = func(self.filename, cache=True)
        self.assertEqual(r, 2)

    def test_cache_reset_mtime(self):
        counter = 0

        @cache.cache
        def func(filename):
            nonlocal counter
            counter +=1
            return counter

        r = func(self.filename)
        self.assertEqual(r, 1)

        t = os.stat(self.filename).st_mtime_ns
        with open(self.filename, 'w') as f:
            f.write('baz')
        os.utime(self.filename, ns=(t, t))

        r = func(self.filename)
        self.assertEqual(r, 1)

    def test_cache_key_sha1(self):
        counter = 0

        @cache.cache(key=utils.sha1sum)
        def func(filename):
            nonlocal counter
            counter +=1
            return counter

        with open(self.filename, 'w') as f:
            f.write('test')

        r = func(self.filename)
        self.assertEqual(r, 1)

        r = func(self.filename)
        self.assertEqual(r, 1)

        t = time.time() + 2.0
        os.utime(self.filename, (t, t))

        r = func(self.filename)
        self.assertEqual(r, 1)

        t = os.stat(self.filename).st_mtime_ns
        with open(self.filename, 'w') as f:
            f.write('baz')
        os.utime(self.filename, ns=(t, t))

        r = func(self.filename)
        self.assertEqual(r, 2)



__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import unittest
import io
import tempfile
import textwrap

from beangulp import csv_utils


class TestCSVUtils(unittest.TestCase):

    def test_csv_clean_header(self):
        self.assertEqual(['date', 'pnl', 'balance'],
                         csv_utils.csv_clean_header('date,p/l,balance'.split(',')))
        self.assertEqual(['date', 'day_s_range', 'balance'],
                         csv_utils.csv_clean_header("date,day's range ,balance".split(',')))
        self.assertEqual(['date', 'col1', 'balance'],
                         csv_utils.csv_clean_header("date,,balance".split(',')))

    def test_csv_dict_reader(self):
        with tempfile.NamedTemporaryFile('w') as tmpfile:
            tmpfile.write(textwrap.dedent(
                """\
                First Name, Last Name, City, Country
                Caroline, Chang, Sydney, Australia
                Martin, Blais, Vancouver, Canada
                """))
            tmpfile.flush()

            with open(tmpfile.name) as infile:
                reader = csv_utils.csv_dict_reader(infile, skipinitialspace=True)
                self.assertTrue(isinstance(reader, object))
                self.assertEqual([
                    {'first_name': 'Caroline',
                     'last_name': 'Chang',
                     'city': 'Sydney',
                     'country': 'Australia'},
                    {'first_name': 'Martin',
                     'last_name': 'Blais',
                     'city': 'Vancouver',
                     'country': 'Canada'}
                    ], list(reader))

    def test_csv_tuple_reader(self):
        with tempfile.NamedTemporaryFile('w') as tmpfile:
            tmpfile.write(textwrap.dedent(
                """\
                First Name, Last Name, City, Country
                Caroline, Chang, Sydney, Australia
                Martin, Blais, Vancouver, Canada
                """))
            tmpfile.flush()

            with open(tmpfile.name) as infile:
                reader = csv_utils.csv_tuple_reader(infile, skipinitialspace=True)
                self.assertTrue(isinstance(reader, object))

                rows = list(reader)
                first_row = rows[0]
                self.assertTrue(isinstance(first_row, tuple))
                self.assertTrue(first_row.first_name)
                self.assertTrue(first_row.last_name)
                self.assertTrue(first_row.city)
                self.assertTrue(first_row.country)
                self.assertEqual(2, len(rows))

    def test_csv_split_sections(self):
        rows = csv_utils.as_rows("""\
        Names:
        First Name, Last Name, City, Country
        Caroline, Chang, Sydney, Australia
        Martin, Blais, Vancouver, Canada

        Ages:
        Last Name, SSN, Age
        Blais, 123-45-6789, 41
        Chang, 987-65-4321, 32
        """)
        sections = csv_utils.csv_split_sections(rows)
        self.assertEqual(
            [[['Names:'],
              ['First Name', ' Last Name', ' City', ' Country'],
              ['Caroline', ' Chang', ' Sydney', ' Australia'],
              ['Martin', ' Blais', ' Vancouver', ' Canada']],
             [['Ages:'],
              ['Last Name', ' SSN', ' Age'],
              ['Blais', ' 123-45-6789', ' 41'],
              ['Chang', ' 987-65-4321', ' 32']]],
            sections)

        rows = csv_utils.as_rows("""\
        Names:
        First Name, Last Name, City, Country
        Caroline, Chang, Sydney, Australia
        Martin, Blais, Vancouver, Canada
        """)
        sections = csv_utils.csv_split_sections(rows)
        self.assertEqual(
            [[['Names:'],
              ['First Name', ' Last Name', ' City', ' Country'],
              ['Caroline', ' Chang', ' Sydney', ' Australia'],
              ['Martin', ' Blais', ' Vancouver', ' Canada']]],
            sections)

        rows = csv_utils.as_rows("")
        sections = csv_utils.csv_split_sections(rows)
        self.assertEqual([], sections)

    def test_csv_split_sections_with_titles(self):
        rows = csv_utils.as_rows("""\
        Names:
        First Name, Last Name, City, Country
        Caroline, Chang, Sydney, Australia
        Martin, Blais, Vancouver, Canada

        Ages:
        Last Name, SSN, Age
        Blais, 123-45-6789, 41
        Chang, 987-65-4321, 32
        """)
        sections = csv_utils.csv_split_sections_with_titles(rows)
        self.assertEqual(
            {'Names:':
             [['First Name', ' Last Name', ' City', ' Country'],
              ['Caroline', ' Chang', ' Sydney', ' Australia'],
              ['Martin', ' Blais', ' Vancouver', ' Canada']],
             'Ages:':
             [['Last Name', ' SSN', ' Age'],
              ['Blais', ' 123-45-6789', ' 41'],
              ['Chang', ' 987-65-4321', ' 32']]},
            sections)

        rows = csv_utils.as_rows("""\
        Names:
        First Name, Last Name, City, Country
        Caroline, Chang, Sydney, Australia
        Martin, Blais, Vancouver, Canada
        """)
        sections = csv_utils.csv_split_sections_with_titles(rows)
        self.assertEqual(
            {'Names:':
             [['First Name', ' Last Name', ' City', ' Country'],
              ['Caroline', ' Chang', ' Sydney', ' Australia'],
              ['Martin', ' Blais', ' Vancouver', ' Canada']]},
            sections)

        rows = csv_utils.as_rows("")
        sections = csv_utils.csv_split_sections_with_titles(rows)
        self.assertEqual({}, sections)


def linearize(iterator, joiner=list):
    """Consume a section iterator.

    Args:
      iterator: An iterator of iterators.
      joiner: A callable to apply to the sub-iterators.
    Returns:
      A list of return values from joiner.
    """
    return list(map(list, iterator))


class TestLineUtils(unittest.TestCase):

    def test_iter_until_empty(self):
        iterator = iter(['a', 'b', '', 'c'])
        prefix = list(csv_utils.iter_until_empty(iterator))
        self.assertEqual(['a', 'b'], prefix)
        self.assertEqual('c', next(iterator))

        iterator = iter(['a', '', '', 'c'])
        prefix = list(csv_utils.iter_until_empty(iterator))
        self.assertEqual(['a'], prefix)
        prefix = list(csv_utils.iter_until_empty(iterator))
        self.assertEqual([], prefix)
        self.assertEqual('c', next(iterator))

    def test_iter_section(self):
        # Test with empty string.
        sio = io.StringIO("")
        self.assertEqual([], linearize(csv_utils.iter_sections(sio)))

        # Test with just empty lines, and a final empty line.
        sio = io.StringIO("\n\n  \n  \n\n  ")
        self.assertEqual([], linearize(csv_utils.iter_sections(sio)))

        # Test with a simple non-empty line.
        sio = io.StringIO("\n\n\n\n\nWORD\n\n\n")
        self.assertEqual([['WORD\n']], linearize(csv_utils.iter_sections(sio)))

        # Test with a simple non-empty line, at the end.
        sio = io.StringIO("\n\n\n\n\nWORD\n")
        self.assertEqual([['WORD\n']], linearize(csv_utils.iter_sections(sio)))

        # Test with a simple non-empty line, at the very end, without a newline.
        sio = io.StringIO("\n\n\n\n\nWORD")
        self.assertEqual([['WORD']], linearize(csv_utils.iter_sections(sio)))

        # Test with output that looks regular.
        sio = io.StringIO("Title1\nA,B,C\nD,E,F\n\n\n\nTitle2\nG,H\nI,J\n,K,L\n\n\n")
        expected = [['Title1\n', 'A,B,C\n', 'D,E,F\n'],
                    ['Title2\n', 'G,H\n', 'I,J\n', ',K,L\n']]
        self.assertEqual(expected, linearize(csv_utils.iter_sections(sio)))


if __name__ == '__main__':
    unittest.main()



"""Code that guesses a MIME type for a filename.

This attempts to identify the mime-type of a file using the built-in
mimetypes library, augmented with MIME types commonly used in
financial downloads.  If this does not produce any match it falls back
to MIME type sniffing using ``python-magic``, if available.

This module is deprecated. Please use ``beancount.mimetypes`` instead.

"""
__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import warnings
from beangulp import mimetypes


# python-magic is an optional dependency.
try:
    import magic
except (ImportError, OSError):
    magic = None


def guess_file_type(filename):
    """Attempt to guess the type of the input file.

    Args:
      filename: A string, the name of the file to guess the type for.
    Returns:
      A suitable mimetype string, or None if we could not guess.
    """

    warnings.warn('beangulp.file_type.guess_file_type() is deprecated. '
                  'Use the beangulp.mimetypes module instead.',
                  DeprecationWarning, stacklevel=2)

    filetype, encoding = mimetypes.guess_type(filename, strict=False)
    if filetype:
        return filetype

    if magic:
        filetype = magic.from_file(filename, mime=True)
        if isinstance(filetype, bytes):
            filetype = filetype.decode('utf8')

    return filetype



from decimal import Decimal
from os import path
from typing import Iterator, Sequence, Union, Set, Optional, Dict
import datetime
import collections
import decimal
import hashlib
import logging
import os
import re

import click

from beangulp import mimetypes


class DefaultDictWithKey(collections.defaultdict):
    """A version of defaultdict whose factory accepts the key as an argument.
    Note: collections.defaultdict would be improved by supporting this directly,
    this is a common occurrence.
    """

    def __missing__(self, key):
        self[key] = value = self.default_factory(key)
        return value


def getmdate(filepath: str) -> datetime.date:
    """Return file modification date."""
    mtime = path.getmtime(filepath)
    return datetime.datetime.fromtimestamp(mtime).date()


def logger(verbosity: int = 0, err: bool = False):
    """Convenient logging method factory."""
    color = False if os.getenv('TERM', '') in ('', 'dumb') else None
    def log(msg, level=0, err=err, **kwargs):
        if level <= verbosity:
            click.secho(msg, color=color, err=err, **kwargs)
    return log


def walk(paths: Sequence[str]) -> Iterator[str]:
    """Yield all the files under paths.

    Takes a sequence of file or directory paths. Directories are
    traversed with os.walk() and complete file paths are returned
    joining filenames to the root directory path. Other elements of
    the list are assumed to be file paths and returned unchanged.

    Args:
      paths: List of filesystems paths.

    Yields:
      Filesystem paths of all the files under paths.

    """
    for file_or_dir in paths:
        if path.isdir(file_or_dir):
            for root, dirs, files in os.walk(file_or_dir):
                for filename in sorted(files):
                    yield path.join(root, filename)
            continue
        yield file_or_dir


def sha1sum(filepath: str) -> str:
    """Compute hash of the file at filepath."""
    with open(filepath, 'rb') as fd:
        return hashlib.sha1(fd.read()).hexdigest()


def is_mimetype(filepath: str, check_mimetypes: Union[str, Set[str]],
                regexp: Optional[bool] = False) -> bool:
    """Check if a file is of one of many mimetypes."""
    if isinstance(check_mimetypes, str):
        check_mimetypes = {check_mimetypes,}
    mtype, _ = mimetypes.guess_type(filepath)
    if mtype is None:
        return False
    return (any(re.fullmatch(r, mtype) for r in check_mimetypes)
            if not regexp else
            (mtype in check_mimetypes))


def search_file_regexp(filepath: str, *regexps: str,
                       nbytes: Optional[int] = None,
                       encoding: Optional[str] = None) -> bool:
    """Check if the header of the file matches the given regexp."""
    with open(filepath, encoding=encoding) as infile:
        # Note: Don't convert just to match on the contents.
        try:
            contents = infile.read(nbytes)
        except UnicodeDecodeError as exc:
            # The encoding wasn't right, don't match.
            logging.warning(f"Error searching for regexp in '{filepath}': {exc}")
            return False
        else:
            return any(re.search(regexp, contents) for regexp in regexps)


def parse_amount(string: str)-> decimal.Decimal:
    """Convert an amount with parens and dollar sign to Decimal."""
    if string is None:
        return Decimal(0)
    string = string.strip()
    if not string:
        return Decimal(0)
    match = re.match(r"\((.*)\)", string)
    if match:
        string = match.group(1)
        sign = -1
    else:
        sign = 1
    cstring = string.replace("-$", "$-").strip(' $').replace(',', '')
    try:
        return Decimal(cstring) * sign
    except decimal.InvalidOperation as exc:
        raise decimal.InvalidOperation(f"Invalid conversion of {cstring!r}") from exc


def validate_accounts(required_accounts: Dict[str, str], provided_accounts: Dict[str, str]):
    """Check a dict of provided account names against a specification of required ones.

    Args:
      required_accounts: A dict of declarations of required values.
      provided_accounts: A config dict of actual values on an importer.
    Raises:
      ValueError: If the configuration is invalid.
    """
    # Note: As an extension, we could provide the existing ledger and try to
    # validate the non-template names against the list of accounts declared in
    # it.

    provided_keys = set(provided_accounts)
    required_keys = set(required_accounts)

    for key in (required_keys - provided_keys):
        raise ValueError(f"Missing value from user configuration: '{key}'; "
                         "against {required_keys}")

    for key in (provided_keys - required_keys):
        raise ValueError(f"Unknown value in user configuration: '{key}'; "
                         f"against {required_keys}")

    for account in provided_accounts.values():
        if not isinstance(account, str):
            raise ValueError(f"Invalid value for account or currency: '{account}'")

def idify(string: str) -> str:
    """Replace characters objectionable for a filename with underscores.
    Args:
      string: Any string.
    Returns:
      The input string, with offending characters replaced.
    """
    for sfrom, sto in [(r"[ \(\)]+", "_"), (r"_*\._*", ".")]:
        string = re.sub(sfrom, sto, string)
    string = string.strip("_")
    return string



__copyright__ = "Copyright (C) 2016,2018  Martin Blais"
__license__ = "GNU GPLv2"

from beangulp import Ingest

def ingest(importers, hooks=None):
    # This function is provided for backward compatibility with the
    # ``beancount.ingest`` framework. It will be removed eventually.
    main = Ingest(importers, hooks)
    main()



"""A file wrapper which acts as a cache for on-demand evaluation of conversions.

This object is used in lieu of a file in order to allow the various importers to
reuse each others' conversion results. Converting file contents, e.g. PDF to
text, can be expensive.
"""
__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import codecs
import functools
import os
import pickle
import sys

from contextlib import suppress
from hashlib import sha1
from os import path

import chardet

from beangulp import mimetypes
from beangulp import utils

# NOTE: See get_file() at the end of this file to create instances of FileMemo.


# Default location of cache directories.
CACHEDIR = (path.expandvars('%LOCALAPPDATA%\\Beangulp')
            if sys.platform == 'win32'
            else path.expanduser('~/.cache/beangulp'))


# Maximum number of bytes to read in order to detect the encoding of a file.
HEAD_DETECT_MAX_BYTES = 128 * 1024


class _FileMemo:
    """A file memoizer which acts as a cache for on-demand evaluation of conversions.

    Attributes:
      name: A string, the name of the underlying file.
    """

    def __init__(self, filename):
        self.name = filename

        # A cache of converter function to saved conversion value.
        self._cache = {}

    def __str__(self):
        return f'<FileWrapper filename="{self.name}">'

    def convert(self, converter_func):
        """Registers a callable used to convert the file contents.
        Args:
          converter_func: A callable which accepts a filename and produces some
          derived version of the file contents.
        Returns:
          A bytes object, with the contents of the entire file.
        """
        try:
            result = self._cache[converter_func]
        except KeyError:
            # FIXME: Implement timing of conversions here. Store it for
            # reporting later.
            result = self._cache[converter_func] = converter_func(self.name)
        return result

    def mimetype(self):
        """Computes the MIME type of the file."""
        return self.convert(mimetype)

    def head(self, num_bytes=8192, encoding=None):
        """An alias for reading just the first bytes of a file."""
        return self.convert(head(num_bytes, encoding=encoding))

    def contents(self):
        """An alias for reading the entire contents of the file."""
        return self.convert(contents)


def mimetype(filename):
    """A converter that computes the MIME type of the file.

    Returns:
      A converter function.
    """
    mtype, _ = mimetypes.guess_type(filename, strict=False)
    return mtype


def head(num_bytes=8192, encoding=None):
    """A converter that just reads the first bytes of a file.

    Note that the returned string may represent less bytes than
    specified if the encoded bytes read from the file terminate with
    an incomplete unicode character. This is likely to occur for
    variable width encodings suach as utf8.

    Args:
      num_bytes: The number of bytes to read.
    Returns:
      A converter function.
    """
    def head_reader(filename):
        with open(filename, 'rb') as fd:
            data = fd.read(num_bytes)
            enc = encoding or chardet.detect(data)['encoding']
            # A little trick to handle an encoded byte array that
            # terminates with an incomplete unicode character.
            decoder = codecs.iterdecode(iter([data]), enc)
            return next(decoder)
    return head_reader


def contents(filename):
    """A converter that just reads the entire contents of a file.

    Args:
      num_bytes: The number of bytes to read.
    Returns:
      A converter function.
    """
    # Attempt to detect the input encoding automatically, using chardet and a
    # decent amount of input.
    with open(filename, 'rb') as infile:
        rawdata = infile.read(HEAD_DETECT_MAX_BYTES)
    detected = chardet.detect(rawdata)
    encoding = detected['encoding']

    # Ignore encoding errors for reading the contents because input files
    # routinely break this assumption.
    errors = 'ignore'

    with open(filename, encoding=encoding, errors=errors) as file:
        return file.read()


def get_file(filename):
    """Create or reuse a globally registered instance of a FileMemo.

    Note: the FileMemo objects' lifetimes are reused for the duration of the
    process. This is usually the intended behavior. Always create them by
    calling this constructor.

    Args:
      filename: A path string, the absolute name of the file whose memo to create.
    Returns:
      A FileMemo instance.

    """
    assert path.isabs(filename), (
        "Path should be absolute in order to guarantee a single call.")
    return _CACHE[filename]


_CACHE = utils.DefaultDictWithKey(_FileMemo)


def cache(func=None, *, key=None):
    """Memoize the result of calling the given function to a disk pickle."""

    def decorator(func):
        @functools.wraps(func)
        def wrapper(filename, *args, cache=None, **kwargs):
            # Compute the cache filename.
            input_key = key(filename) if key else filename
            name = sha1(pickle.dumps((input_key, args, kwargs))).hexdigest() + '.pickle'
            cache_fname = path.join(CACHEDIR, name)

            # We inspect the modified time of the input file and the cache.
            input_mtime = os.stat(filename).st_mtime_ns
            cache_mtime = 0
            with suppress(FileNotFoundError):
                cache_mtime = os.stat(cache_fname).st_mtime_ns

            if cache is None:
                # Read from cache when a key function has been supplied and the
                # cache file exists or when the filename has been used to
                # compute the cache key and the cache entry modification time is
                # equal or later the input file modification time.
                cache = cache_mtime != 0 if key else cache_mtime >= input_mtime

            if cache:
                with open(cache_fname, 'rb') as f:
                    return pickle.load(f)

            # Invoke the potentially expensive function.
            ret = func(filename, *args, **kwargs)

            # Ignore errors due to the CACHEDIR not being present.
            with suppress(FileNotFoundError):
                # To populate the cache atomically write the cache entry in a
                # temporary file and move it to the right place with the
                # complete content and the right modification time.
                cache_temp = cache_fname + '~'
                with open(cache_temp, 'wb') as f:
                    pickle.dump(ret, f)
                os.utime(cache_temp, ns=(input_mtime, input_mtime))
                os.replace(cache_temp, cache_fname)

            return ret
        return wrapper

    if func is None:
        return decorator
    return decorator(func)



"""Example importer for example broker UTrade.
"""
__copyright__ = "Copyright (C) 2016  Martin Blais"
__license__ = "GNU GPLv2"

import csv
import datetime
import re
import logging

from os import path
from dateutil.parser import parse

from beancount.core import account
from beancount.core import amount
from beancount.core import data
from beancount.core import flags
from beancount.core import position
from beancount.core.number import D
from beancount.core.number import ZERO

import beangulp
from beangulp.testing import main


class Importer(beangulp.Importer):
    """An importer for UTrade CSV files (an example investment bank)."""

    def __init__(self, currency,
                 account_root,
                 account_cash,
                 account_dividends,
                 account_gains,
                 account_fees,
                 account_external):
        self.currency = currency
        self.account_root = account_root
        self.account_cash = account_cash
        self.account_dividends = account_dividends
        self.account_gains = account_gains
        self.account_fees = account_fees
        self.account_external = account_external

    def identify(self, filepath):
        # Match if the filename is as downloaded and the header has the unique
        # fields combination we're looking for.
        if not re.match(r"UTrade\d\d\d\d\d\d\d\d\.csv", path.basename(filepath)):
            return False
        with open(filepath, 'r') as fd:
            head = fd.read(13)
        if head != "DATE,TYPE,REF":
            return False
        return True

    def filename(self, filepath):
        return 'utrade.{}'.format(path.basename(filepath))

    def account(self, filepath):
        return self.account_root

    def date(self, filepath):
        # Extract the statement date from the filename.
        return datetime.datetime.strptime(path.basename(filepath),
                                          'UTrade%Y%m%d.csv').date()

    def extract(self, filepath, existing):
        # Open the CSV file and create directives.
        entries = []
        index = 0
        with open(filepath) as infile:
            for index, row in enumerate(csv.DictReader(infile)):
                meta = data.new_metadata(filepath, index)
                date = parse(row['DATE']).date()
                rtype = row['TYPE']
                link = f"ut{row['REF #']}"
                desc = f"({row['TYPE']}) {row['DESCRIPTION']}"
                units = amount.Amount(D(row['AMOUNT']), self.currency)
                fees = amount.Amount(D(row['FEES']), self.currency)
                other = amount.add(units, fees)

                if rtype == 'XFER':
                    assert fees.number == ZERO
                    txn = data.Transaction(
                        meta, date, flags.FLAG_OKAY, None, desc, data.EMPTY_SET, {link}, [
                            data.Posting(self.account_cash, units, None, None, None,
                                         None),
                            data.Posting(self.account_external, -other, None, None, None,
                                         None),
                        ])

                elif rtype == 'DIV':
                    assert fees.number == ZERO

                    # Extract the instrument name from its description.
                    match = re.search(r'~([A-Z]+)$', row['DESCRIPTION'])
                    if not match:
                        logging.error("Missing instrument name in '%s'", row['DESCRIPTION'])
                        continue
                    instrument = match.group(1)
                    account_dividends = self.account_dividends.format(instrument)

                    txn = data.Transaction(
                        meta, date, flags.FLAG_OKAY, None, desc, data.EMPTY_SET, {link}, [
                            data.Posting(self.account_cash, units, None, None, None, None),
                            data.Posting(account_dividends, -other, None, None, None, None),
                        ])

                elif rtype in ('BUY', 'SELL'):

                    # Extract the instrument name, number of units, and price from
                    # the description. That's just what we're provided with (this is
                    # actually realistic of some data from some institutions, you
                    # have to figure out a way in your parser).
                    match = re.search(r'\+([A-Z]+)\b +([0-9.]+)\b +@([0-9.]+)',
                                      row['DESCRIPTION'])
                    if not match:
                        logging.error("Missing purchase infos in '%s'", row['DESCRIPTION'])
                        continue
                    instrument = match.group(1)
                    account_inst = account.join(self.account_root, instrument)
                    units_inst = amount.Amount(D(match.group(2)), instrument)
                    rate = D(match.group(3))

                    if rtype == 'BUY':
                        cost = position.Cost(rate, self.currency, None, None)
                        txn = data.Transaction(
                            meta, date, flags.FLAG_OKAY, None, desc, data.EMPTY_SET, {link}, [
                                data.Posting(self.account_cash, units, None, None, None,
                                             None),
                                data.Posting(self.account_fees, fees, None, None, None,
                                             None),
                                data.Posting(account_inst, units_inst, cost, None, None,
                                             None),
                            ])

                    elif rtype == 'SELL':
                        # Extract the lot. In practice this information not be there
                        # and you will have to identify the lots manually by editing
                        # the resulting output. You can leave the cost.number slot
                        # set to None if you like.
                        match = re.search(r'\(LOT ([0-9.]+)\)', row['DESCRIPTION'])
                        if not match:
                            logging.error("Missing cost basis in '%s'", row['DESCRIPTION'])
                            continue
                        cost_number = D(match.group(1))
                        cost = position.Cost(cost_number, self.currency, None, None)
                        price = amount.Amount(rate, self.currency)
                        account_gains = self.account_gains.format(instrument)
                        txn = data.Transaction(
                            meta, date, flags.FLAG_OKAY, None, desc, data.EMPTY_SET, {link}, [
                                data.Posting(self.account_cash, units, None, None, None,
                                             None),
                                data.Posting(self.account_fees, fees, None, None, None,
                                             None),
                                data.Posting(account_inst, units_inst, cost, price, None,
                                             None),
                                data.Posting(account_gains, None, None, None, None,
                                             None),
                            ])

                else:
                    logging.error("Unknown row type: %s; skipping", rtype)
                    continue

                entries.append(txn)

        # Insert a final balance check.
        if index:
            entries.append(
                data.Balance(meta, date + datetime.timedelta(days=1),
                             self.account_cash,
                             amount.Amount(D(row['BALANCE']), self.currency),
                             None, None))

        return entries

    @staticmethod
    def cmp(a, b):
        # This importer attaches an unique ID to all transactions in
        # the form of a link. The link can be used to implement
        # transaction duplicates detection based on transactions IDs.

        if not isinstance(a, data.Transaction):
            return False
        if not isinstance(b, data.Transaction):
            return False

        # Get all the links with the expected ut$ID format.
        aids = [link for link in a.links if re.match(r'ut\d{8}', link)]
        if not aids:
            # If there are no matching links, stop here.
            return False

        # Get all the links with the expected ut$ID format.
        bids = [link for link in b.links if re.match(r'ut\d{8}', link)]
        if not bids:
            # If there are no matching links, stop here.
            return False

        if len(aids) != len(bids):
            return False

        # Compare all collected IDs.
        if all(aid == bid for aid, bid in zip(sorted(aids), sorted(bids))):
            return True

        return False


if __name__ == '__main__':
    importer = Importer(
        "USD",
        "Assets:US:UTrade",
        "Assets:US:UTrade:Cash",
        "Income:US:UTrade:{}:Dividend",
        "Income:US:UTrade:{}:Gains",
        "Expenses:Financial:Fees",
        "Assets:US:BofA:Checking")
    main(importer)



# TODO(blais): Make this only register the new types, and keep the user
# referring to the interface directly. There's no real need to 'import *' here.
# The root package would automatically register those new types.

"""The standard Python 'mimetypes' module with some custom registrations for
types commonly used for accounting. Import 'from beangulp import mimetypes'
instead of 'import mimetypes'.
"""

from mimetypes import *  # noqa: F403

# Register some MIME types used in financial downloads.
_extra_mime_types = [
    ('text/beancount', '.beancount', '.beans'), # Beancount ledgers.
    ('application/vnd.intu.qbo', '.qbo'), # Quicken files.
    ('application/x-ofx', '.qfx', '.ofx'), # Open Financial Exchange files.
]

for mime, *extensions in _extra_mime_types:
    for ext in extensions:
        add_type(mime, ext, strict=False)  # noqa: F405



import contextlib
import textwrap
import traceback


class Error(RuntimeError):
    def __init__(self, message, *args):
        self.message = message
        self.args = args

    def __str__(self):
        return '\n'.join((*self.args, self.message))


class ExceptionsTrap(contextlib.AbstractContextManager):
    """A context manager to log exceptions.

    Works similarly to contextlib.suppress() but logs the exceptions
    instead than simply discarding them. This is used to shorten and
    unify exception handling in the command line wrappers. The format
    of the reporting is specific to the Beangulp command line
    interface.

    """
    def __init__(self, func):
        self.log = func
        self.errors = 0

    def __exit__(self, exctype, excinst, exctb):
        if exctype is None:
            return True
        self.errors += 1
        self.log('  ERROR', fg='red')
        if issubclass(exctype, Error):
            # Beangulp validation error.
            self.log(textwrap.indent(str(excinst), '  '))
            return True
        if issubclass(exctype, Exception):
            # Unexpected exception.
            self.log('  Exception in importer code.')
            exc = ''.join(traceback.format_exception(exctype, excinst, exctb))
            self.log(textwrap.indent(exc, '  ').rstrip())
            return True
        return False

    def __bool__(self):
        """Return True if any error occurred."""
        return self.errors != 0
